import chess
from .types import *
from .player import Player
from .game import Game, LocalGame, RemoteGame
from .history import GameHistory


def play_local_game(white_player: Player, black_player: Player, game: LocalGame = None,
                    seconds_per_player: float = 900) -> Tuple[Optional[Color], Optional[WinReason], GameHistory]:
    """
    Plays a game between the two players passed in. Uses :class:`LocalGame` to run the game, and just calls
    :func:`play_turn` until the game is over: ::

        while not game.is_over():
            play_turn(game, player)

    :param white_player: The white :class:`Player`.
    :param black_player: The black :class:`Player`.
    :param game: The :class:`LocalGame` object to use.
    :param seconds_per_player: The time each player has to play. Only used if `game` is not passed in.
    :return: The results of the game, also passed to each player via :meth:`Player.handle_game_end`.
    """
    players = [black_player, white_player]

    if game is None:
        game = LocalGame(seconds_per_player=seconds_per_player)

    white_name = white_player.__class__.__name__
    black_name = black_player.__class__.__name__
    game.store_players(white_name, black_name)

    white_player.handle_game_start(chess.WHITE, game.board.copy(), black_name)
    black_player.handle_game_start(chess.BLACK, game.board.copy(), white_name)
    game.start()

    while not game.is_over():
        play_turn(game, players[game.turn], end_turn_last=True)

    game.end()
    winner_color = game.get_winner_color()
    win_reason = game.get_win_reason()
    game_history = game.get_game_history()

    white_player.handle_game_end(winner_color, win_reason, game_history)
    black_player.handle_game_end(winner_color, win_reason, game_history)

    return winner_color, win_reason, game_history


def play_remote_game(server_url, game_id, auth, player: Player):
    game = RemoteGame(server_url, game_id, auth)

    player.handle_game_start(game.get_player_color(), game.get_starting_board(), game.get_opponent_name())
    game.start()

    while not game.is_over():
        play_turn(game, player, end_turn_last=False)

    winner_color = game.get_winner_color()
    win_reason = game.get_win_reason()
    game_history = game.get_game_history()

    player.handle_game_end(winner_color, win_reason, game_history)

    return winner_color, win_reason, game_history


def play_turn(game: Game, player: Player, end_turn_last=False):
    """
    Coordinates playing a turn for `player` in `game`. Does the following sequentially:

    #. :func:`notify_opponent_move_results`
    #. :func:`play_sense`
    #. :func:`play_move`

    See :func:`play_move` for more info on `end_turn_last`.

    :param game: The :class:`Game` that `player` is playing in.
    :param player: The :class:`Player` whose turn it is.
    :param end_turn_last: Flag indicating whether to call :meth:`Game.end_turn` before or after :meth:`Player.handle_move_result`
    """
    sense_actions = game.sense_actions()
    move_actions = game.move_actions()

    notify_opponent_move_results(game, player)

    play_sense(game, player, sense_actions, move_actions)

    play_move(game, player, move_actions, end_turn_last=end_turn_last)


def notify_opponent_move_results(game: Game, player: Player):
    """
    Passes the opponents move results to the player. Does the following sequentially:

    #. Get the results of the opponents move using :meth:`Game.opponent_move_results`.
    #. Give the results to the player using :meth:`Player.handle_opponent_move_result`.

    :param game: The :class:`Game` that `player` is playing in.
    :param player: The :class:`Player` whose turn it is.
    """
    opt_capture_square = game.opponent_move_results()
    player.handle_opponent_move_result(opt_capture_square is not None, opt_capture_square)


def play_sense(game: Game, player: Player, sense_actions: List[Square], move_actions: List[chess.Move]):
    """
    Runs the sense phase for `player` in `game`. Does the following sequentially:

    #. Get the sensing action using :meth:`Player.choose_sense`.
    #. Apply the sense action using :meth:`Game.sense`.
    #. Give the result of the sense action to player using :meth:`Player.handle_sense_result`.

    :param game: The :class:`Game` that `player` is playing in.
    :param player: The :class:`Player` whose turn it is.
    :param sense_actions: The possible sense actions for `player`.
    :param move_actions: The possible move actions for `player`.
    """
    sense = player.choose_sense(sense_actions, move_actions, game.get_seconds_left())
    sense_result = game.sense(sense)
    player.handle_sense_result(sense_result)


def play_move(game: Game, player: Player, move_actions: List[chess.Move], end_turn_last=False):
    """
    Runs the move phase for `player` in `game`. Does the following sequentially:

    #. Get the moving action using :meth:`Player.choose_move`.
    #. Apply the moving action using :meth:`Game.move`.
    #. Ends the current player's turn using :meth:`Game.end_turn`.
    #. Give the result of the moveaction to player using :meth:`Player.handle_move_result`.

    If `end_turn_last` is True, then :meth:`Game.end_turn` is called last instead of before
    :meth:`Player.handle_move_result`.

    :param game: The :class:`Game` that `player` is playing in.
    :param player: The :class:`Player` whose turn it is.
    :param move_actions: The possible move actions for `player`.
    :param end_turn_last: Flag indicating whether to call :meth:`Game.end_turn` before or after :meth:`Player.handle_move_result`
    """
    move = player.choose_move(move_actions, game.get_seconds_left())
    requested_move, taken_move, opt_enemy_capture_square = game.move(move)

    if not end_turn_last:
        game.end_turn()

    player.handle_move_result(requested_move, taken_move,
                              opt_enemy_capture_square is not None, opt_enemy_capture_square)

    if end_turn_last:
        game.end_turn()



from abc import abstractmethod
from datetime import datetime
import math
from typing import Optional

import chess
import requests
import time
from .utilities import *
from .history import GameHistory, GameHistoryDecoder


class Game(object):
    """
    Abstract class that represents an instantiation of a Reconnaissance Chess Game. See :class:`LocalGame`
    and :class:`RemoteGame`
    for implementations.
    """

    @abstractmethod
    def sense_actions(self) -> List[Square]:
        """
        :return: List of :class:`Square` that the player can choose for sense phase.
        """
        pass

    @abstractmethod
    def move_actions(self) -> List[chess.Move]:
        """
        :return: List of :class:`chess.Move` that the player can choose for move phase.
        """
        pass

    @abstractmethod
    def get_seconds_left(self) -> float:
        """
        :return: float indicating the seconds the player has left to play.
        """
        pass

    @abstractmethod
    def start(self):
        """
        Starts the game and the timers for each player.
        """
        pass

    @abstractmethod
    def opponent_move_results(self) -> Optional[Square]:
        """
        :return: :class:`Square` where opponent captured a piece last turn if they did, otherwise `None`.
        """
        pass

    @abstractmethod
    def sense(self, square: Optional[Square]) -> List[Tuple[Square, Optional[chess.Piece]]]:
        """
        Execute a sense action and get the sense result.

        An example result returned from sensing `B7` at the beginning of the game: ::

            [
                (A8, Piece(ROOK, BLACK)), (B8, Piece(KNIGHT, BLACK)), (C8, Piece(BISHOP, BLACK)),
                (A7, Piece(PAWN, BLACK)), (B7, Piece(PAWN, BLACK)), (C7, Piece(PAWN, BLACK)),
                (A6, None), (B6, None), (C8, None)
            ]

        :param square: The :class:`Square` to sense.
        :return: A list of tuples, where each tuple contains a :class:`Square` in the sense, and if there was a piece
            on the square, then the corresponding :class:`chess.Piece`, otherwise `None`.
        """
        pass

    @abstractmethod
    def move(self, requested_move: Optional[chess.Move]) \
            -> Tuple[Optional[chess.Move], Optional[chess.Move], Optional[Square]]:
        """
        Execute a move action and get the result.

        :param requested_move: The :class:`chess.Move` to execute.
        :return: A tuple containing the requested :class:`chess.Move`, the taken :class:`chess.Move`,
            and the :class:`Square` that a capture occurred on if one occurred.
        """
        pass

    @abstractmethod
    def end_turn(self):
        """
        End the current player's turn.
        """
        pass

    @abstractmethod
    def is_over(self) -> bool:
        """
        The game is over if either player has run out of time, or if either player's King has been captured.

        This will always return `True` after :meth:`end()` has been called.

        :return: Returns `True` if the game is over, otherwise `False`.
        """
        pass

    @abstractmethod
    def get_winner_color(self) -> Optional[Color]:
        """
        Returns the color of the player who won the game. If the game is not over, or is over but does not have a winner
        (i.e. :meth:`end()` has been called), then this will return `None`.

        :return: :class:`Color` of the winner if the game has ended and has a winner, otherwise `None`.
        """
        pass

    @abstractmethod
    def get_win_reason(self) -> Optional[WinReason]:
        """
        Returns the reason the player who won won the game. If the game is not over, or is over but does not have a
        winner (i.e. :meth:`end()` has been called), then this will return `None`.

        :return: :class:`WinReason` of the winner if the game has ended and has a winner, otherwise `None`.
        """
        pass

    @abstractmethod
    def get_game_history(self) -> Optional[GameHistory]:
        """
        Get the history of the game.

        :return: :class:`GameHistory` if the game is over, otherwise `None`.
        """
        pass


class LocalGame(Game):
    """
    The local implementation of :class:`Game`. Used to run games locally instead of remotely via a server.
    """

    def __init__(
            self,
            seconds_per_player: Optional[float] = 900,
            seconds_increment: Optional[float] = 5,
            reversible_moves_limit: Optional[int] = 100,
            full_turn_limit: Optional[int] = None,
    ):
        """
        Constructs the Game object

        :param seconds_per_player: Initial clock limit for each player in seconds.
            Use None for unlimited. Default is 900.
        :param seconds_increment: Seconds added to a player's clock limit on each turn.
            Will treat None as 0. Default is 5.
        :param reversible_moves_limit: Maximum number of moves without pawn moves or captures before game is a draw.
            Use None for unlimited. Default is 100 (a non-optional version of the "`50-move rule`_").
        :param full_turn_limit: Maximum number of full turns (both players move) before game is a draw.
            Use None for unlimited. Default is None.

        .. _50-move rule: https://en.wikipedia.org/wiki/Fifty-move_rule
        """
        self.seconds_per_player = seconds_per_player if seconds_per_player is not None else math.inf
        self.seconds_increment = seconds_increment if seconds_increment is not None else 0
        self.reversible_moves_limit = reversible_moves_limit if reversible_moves_limit is not None else math.inf
        self.full_turn_limit = full_turn_limit if full_turn_limit is not None else math.inf

        self.turn = chess.WHITE
        self.board = chess.Board()

        self.__game_history = GameHistory()

        self._is_finished = False
        self._resignee = None
        self.seconds_left_by_color = {
            chess.WHITE: self.seconds_per_player,
            chess.BLACK: self.seconds_per_player,
        }
        self.current_turn_start_time = None

        self.move_results = None

    def start(self):
        """
        Starts off the clock for the first player.

        :return: None.
        """
        self.current_turn_start_time = datetime.now()

    def end(self):
        """
        Ends the game.
        :return: None.
        """
        self.seconds_left_by_color[self.turn] = self.get_seconds_left()
        self._is_finished = True
        self.__game_history.store_results(self.get_winner_color(), self.get_win_reason())

    def store_players(self, white_name, black_name):
        self.__game_history.store_players(white_name, black_name)

    def resign(self):
        self._resignee = self.turn

    def get_seconds_left(self) -> float:
        """
        :return: The amount of seconds left for the current player.
        """
        if not self._is_finished and self.current_turn_start_time:
            elapsed_since_turn_start = (datetime.now() - self.current_turn_start_time).total_seconds()
            return self.seconds_left_by_color[self.turn] - elapsed_since_turn_start
        else:
            return self.seconds_left_by_color[self.turn]

    def sense_actions(self) -> List[Square]:
        """
        :return: List of all squares on the board.
        """
        return None if self._is_finished else list(chess.SQUARES)

    def move_actions(self) -> List[chess.Move]:
        """
        :return: List of moves that are possible with only knowledge of your pieces
        """
        return None if self._is_finished else move_actions(self.board)

    def opponent_move_results(self) -> Optional[Square]:
        return self.move_results

    def sense(self, square: Optional[Square]) -> List[Tuple[Square, Optional[chess.Piece]]]:
        if self._is_finished:
            return []

        if square is None:
            # don't sense anything
            sense_result = []
        else:
            if square not in self.sense_actions():
                raise ValueError('LocalGame::sense({}): {} is not a valid square.'.format(square, square))

            rank, file = chess.square_rank(square), chess.square_file(square)
            sense_result = []
            for delta_rank in [1, 0, -1]:
                for delta_file in [-1, 0, 1]:
                    if 0 <= rank + delta_rank <= 7 and 0 <= file + delta_file <= 7:
                        sense_square = chess.square(file + delta_file, rank + delta_rank)
                        sense_result.append((sense_square, self.board.piece_at(sense_square)))

        self.__game_history.store_sense(self.turn, square, sense_result)

        return sense_result

    def move(self, requested_move: Optional[chess.Move]) \
            -> Tuple[Optional[chess.Move], Optional[chess.Move], Optional[Square]]:
        if self._is_finished:
            return requested_move, None, None

        if requested_move is None:
            # pass move
            taken_move = None

            # doesn't capture anything
            opt_capture_square = None
        else:
            # add in a queen promotion if the move doesn't have one but could have one
            move = add_pawn_queen_promotion(self.board, requested_move)
            if move not in self.move_actions():
                raise ValueError('Requested move {} was not in move_actions()'.format(requested_move))

            # calculate taken move
            taken_move = self._revise_move(move)

            # calculate capture square
            opt_capture_square = capture_square_of_move(self.board, taken_move)

        # store move information before the move is pushed, as pushing a move
        # will change the turn over to the opponent
        self.__game_history.store_move(self.turn, requested_move, taken_move, opt_capture_square)
        self.__game_history.store_fen_before_move(self.turn, self.board.fen(en_passant='fen'))

        # apply move
        self.board.push(taken_move if taken_move is not None else chess.Move.null())

        self.__game_history.store_fen_after_move(self.turn, self.board.fen(en_passant='fen'))

        # store results of move for notifying other player
        self.move_results = opt_capture_square

        return requested_move, taken_move, opt_capture_square

    def _revise_move(self, move):
        return revise_move(self.board, move)

    def end_turn(self):
        """
        Used for bookkeeping. Does the following:

        #. Updates the time used for the current player
        #. Ends the turn for the current player
        #. Starts the timer for the next player

        :return: None
        """
        elapsed = datetime.now() - self.current_turn_start_time
        self.seconds_left_by_color[self.turn] -= elapsed.total_seconds()
        self.seconds_left_by_color[self.turn] += self.seconds_increment

        self.turn = not self.turn
        self.current_turn_start_time = datetime.now()

    def get_game_history(self) -> Optional[GameHistory]:
        return self.__game_history if self.is_over() else None

    def is_over(self) -> bool:
        if self._is_finished:
            return True

        no_time_left = self.get_seconds_left() <= 0 or self.seconds_left_by_color[not self.turn] <= 0
        king_captured = self.board.king(chess.WHITE) is None or self.board.king(chess.BLACK) is None
        someone_resigned = self._resignee is not None

        max_turns_reached = self.board.fullmove_number > self.full_turn_limit
        move_limit_reached = self.board.halfmove_clock >= self.reversible_moves_limit

        return no_time_left or king_captured or someone_resigned or max_turns_reached or move_limit_reached

    def get_winner_color(self) -> Optional[Color]:
        if not self.is_over():
            return None

        if self._resignee is not None:
            return not self._resignee

        if self.seconds_left_by_color[chess.WHITE] <= 0:
            return chess.BLACK
        elif self.seconds_left_by_color[chess.BLACK] <= 0:
            return chess.WHITE

        if self.board.king(chess.WHITE) is None:
            return chess.BLACK
        elif self.board.king(chess.BLACK) is None:
            return chess.WHITE

        return None

    def get_win_reason(self) -> Optional[WinReason]:
        if not self.is_over():
            return None

        if self._resignee is not None:
            return WinReason.RESIGN
        elif self.seconds_left_by_color[chess.WHITE] <= 0 or self.seconds_left_by_color[chess.BLACK] <= 0:
            return WinReason.TIMEOUT
        elif self.board.king(chess.WHITE) is None or self.board.king(chess.BLACK) is None:
            return WinReason.KING_CAPTURE
        elif self.board.fullmove_number > self.full_turn_limit:
            return WinReason.TURN_LIMIT
        elif self.board.halfmove_clock >= self.reversible_moves_limit:
            return WinReason.MOVE_LIMIT

        return None


class RemoteGame(Game):
    """
    The remote implementation of :class:`Game`. Used to play games remotely via a server.

    All the methods implemented are pass-throughs to the server. Each method submits a HTTP request to the corresponding
    end point on the server.
    """

    def __init__(self, server_url, game_id, auth):
        self.game_url = '{}/api/games/{}'.format(server_url, game_id)
        self.session = requests.Session()
        self.session.auth = auth

    def _get(self, endpoint, decoder_cls=ChessJSONDecoder):
        url = '{}/{}'.format(self.game_url, endpoint)
        while True:
            try:
                response = self.session.get(url)
                if response.status_code == 200:
                    return response.json(cls=decoder_cls)
                elif response.status_code >= 500:
                    time.sleep(0.5)
                else:
                    raise ValueError(response.text)
            except requests.RequestException as e:
                print(e)
                time.sleep(0.5)

    def _post(self, endpoint, obj):
        url = '{}/{}'.format(self.game_url, endpoint)
        data = json.dumps(obj, cls=ChessJSONEncoder)
        while True:
            try:
                response = self.session.post(url, data=data)
                if response.status_code == 200:
                    return response.json(cls=ChessJSONDecoder)
                elif response.status_code >= 500:
                    time.sleep(0.5)
                else:
                    raise ValueError(response.text)
            except requests.RequestException as e:
                print(e)
                time.sleep(0.5)

    def get_player_color(self):
        return self._get('color')['color']

    def get_starting_board(self):
        return self._get('starting_board')['board']

    def get_opponent_name(self):
        return self._get('opponent_name')['opponent_name']

    def sense_actions(self) -> List[Square]:
        return self._get('sense_actions')['sense_actions']

    def move_actions(self) -> List[chess.Move]:
        return self._get('move_actions')['move_actions']

    def get_seconds_left(self) -> float:
        return self._get('seconds_left')['seconds_left']

    def start(self):
        self._post('ready', {})

    def is_my_turn(self) -> bool:
        return self._get('is_my_turn')['is_my_turn']

    def opponent_move_results(self) -> Optional[Square]:
        return self._get('opponent_move_results')['opponent_move_results']

    def sense(self, square: Optional[Square]) -> List[Tuple[Square, Optional[chess.Piece]]]:
        return self._post('sense', {'square': square})['sense_result']

    def move(self, requested_move: Optional[chess.Move]) -> Tuple[
        Optional[chess.Move], Optional[chess.Move], Optional[Square]]:
        return self._post('move', {'requested_move': requested_move})['move_result']

    def end_turn(self):
        self._post('end_turn', {})

    def is_over(self) -> bool:
        while True:
            status = self._get('game_status')
            if status['is_over']:
                return True
            if status['is_my_turn']:
                return False
            time.sleep(0.5)

    def get_winner_color(self) -> Optional[Color]:
        return self._get('winner_color')['winner_color']

    def get_win_reason(self) -> Optional[WinReason]:
        return self._get('win_reason')['win_reason']

    def get_game_history(self) -> Optional[GameHistory]:
        return self._get('game_history', decoder_cls=GameHistoryDecoder)['game_history']



import chess
from .types import *
from typing import Callable, TypeVar, Iterable, Mapping
import json
import math
from .utilities import ChessJSONEncoder, ChessJSONDecoder

T = TypeVar('T')


class Turn(object):
    """
    The representation of a single turn in a game. Contains the color of the player who played this turn, as well
    as the number of turns the player has taken so far.
    """

    def __init__(self, color: Color, turn_number: int):
        self.color = color
        self.turn_number = turn_number

    @property
    def next(self):
        """
        :return: The :class:`Turn` that happens immediately after this, which is the other player's next turn.
        """
        return Turn(not self.color, self.turn_number + (0 if self.color == chess.WHITE else 1))

    @property
    def previous(self):
        """
        :return: The :class: `Turn` that happens immediately before this, which is the other player's previous turn.
        """
        return Turn(not self.color, self.turn_number - (1 if self.color == chess.WHITE else 0))

    def __eq__(self, other):
        if not isinstance(other, Turn):
            return NotImplemented

        return self.color == other.color and self.turn_number == other.turn_number

    def __lt__(self, other):
        if not isinstance(other, Turn):
            return NotImplemented

        if self.turn_number < other.turn_number:
            return True
        elif self.turn_number > other.turn_number:
            return False
        else:
            return self.color == chess.WHITE and other.color == chess.BLACK

    def __le__(self, other):
        if not isinstance(other, Turn):
            return NotImplemented

        return self == other or self < other

    def __str__(self):
        return 'Turn({}, {})'.format(chess.COLOR_NAMES[self.color], self.turn_number)

    def __repr__(self):
        return str(self)


class GameHistory(object):
    """
    Implements method for processing and querying a Game.

    Here are some example uses:

    Extracting data to train a sensing policy to sense opponent: ::

        opponent_moves = history.collect(history.taken_move, history.turns(not self.color))
        target_senses = [move.to_square for move in opponent_moves]

    Extracting data to train a movement policy: ::

        for turn in history.turns(self.color):
            if history.has_move(turn):
                move = history.requested_move(turn)
                board = history.truth_board_before_move(turn)
                # do something with move and board...

    Randomly sampling from the turns instead of using them in sequential order: ::

        for turn in random.sample(history.turns(self.color), N):
            ...

    Giving rewards based on validity of the move: ::

        for turn in history.turns(self.color):
            if history.has_move(turn):
                if history.requested_move(turn) != history.taken_move(turn):
                    reward = -1
                else:
                    reward = 1

    Giving rewards for moving out of check: ::

        for turn in history.turns(self.color):
            if history.has_move(turn):
                board_before = history.truth_board_before_move(turn)
                board_after = history.truth_board_after_move(turn)

                if board_before.is_check() and not board_after.is_check()
                    reward = 1
                else:
                    reward = -1

    """

    def __init__(self):
        self._white_name = None
        self._black_name = None
        self._senses = {chess.WHITE: [], chess.BLACK: []}
        self._sense_results = {chess.WHITE: [], chess.BLACK: []}
        self._requested_moves = {chess.WHITE: [], chess.BLACK: []}
        self._taken_moves = {chess.WHITE: [], chess.BLACK: []}
        self._capture_squares = {chess.WHITE: [], chess.BLACK: []}
        self._fens_before_move = {chess.WHITE: [], chess.BLACK: []}
        self._fens_after_move = {chess.WHITE: [], chess.BLACK: []}
        self._winner_color = None
        self._win_reason = None

    def save(self, filename):
        """
        Save the game history to a json file.

        :param filename: The file to save to.
        """
        with open(filename, 'w', newline='') as fp:
            json.dump(self, fp, cls=GameHistoryEncoder)

    @classmethod
    def from_file(cls, filename):
        """
        :param filename: The json file to load the :class:`GameHistory` object from.
        :return: The :class:`GameHistory` object that was originally saved to the file using :meth:`save`.
        """
        with open(filename, newline='') as fp:
            return json.load(fp, cls=GameHistoryDecoder)

    def store_players(self, white_name: str, black_name: str):
        self._white_name = white_name
        self._black_name = black_name

    def store_sense(self, color: Color, square: Optional[Square],
                    sense_result: List[Tuple[Square, Optional[chess.Piece]]]):
        self._senses[color].append(square)
        self._sense_results[color].append(sense_result)

    def store_move(self, color: Color, requested_move: Optional[chess.Move],
                   taken_move: Optional[chess.Move], opt_capture_square: Optional[Square]):
        self._requested_moves[color].append(requested_move)
        self._taken_moves[color].append(taken_move)
        self._capture_squares[color].append(opt_capture_square)

    def store_fen_before_move(self, color: Color, fen: str):
        self._fens_before_move[color].append(fen)

    def store_fen_after_move(self, color: Color, fen: str):
        self._fens_after_move[color].append(fen)

    def store_results(self, winner_color: Optional[Color], win_reason: Optional[WinReason]):
        self._winner_color = winner_color
        self._win_reason = win_reason

    def get_white_player_name(self) -> str:
        """
        Get the name of white.
        :return: str
        """
        return self._white_name

    def get_black_player_name(self) -> str:
        """
        Get the name of black.
        :return: str
        """
        return self._black_name

    def is_empty(self) -> bool:
        """
        Get whether or not the game had any turns in it.

        Examples:
            >>> history.is_empty()
            False

        :return: `True` if there are no turns to query in this object, `False` otherwise.
        """
        return len(self._senses[chess.WHITE]) == 0

    def num_turns(self, color: Color = None) -> int:
        """
        Get the number of turns taken in the game. Optionally specify the color of the player to get the number of turns
        for.

        Examples:
            >>> history.num_turns()
            9

            >>> history.num_turns(WHITE)
            5

            >>> history.num_turns(BLACK)
            4

        :param color: Optional player color indicating which player's number of turns to return.
        :return: The number of turns saved in this object. If `color` is specified, get the number of turns for that
            player.
        """
        return len(list(self.turns(color=color)))

    def turns(self, color: Color = None, start=0, stop=math.inf) -> Iterable[Turn]:
        """
        Get all the turns that happened in the game in order. Optionally specify a single player to get only that
        player's turns.

        Examples:
            >>> list(history.turns())
            [Turn(WHITE, 0), Turn(BLACK, 0), Turn(WHITE, 1), ..., Turn(BLACK, 23)]

            >>> list(history.turns(WHITE))
            [Turn(WHITE, 0), Turn(WHITE, 1), ..., Turn(WHITE, 22)]

            >>> list(history.turns(BLACK))
            [Turn(BLACK, 0), Turn(BLACK, 1), ..., Turn(BLACK, 23)]

            >>> list(history.turns(start=1))
            [Turn(WHITE, 1), Turn(BLACK, 1), Turn(WHITE, 2), ..., Turn(BLACK, 23)]

            >>> list(history.turns(stop=2))
            [Turn(WHITE, 0), Turn(BLACK, 0), Turn(WHITE, 1), Turn(BLACK, 1)]

            >>> list(history.turns(WHITE, stop=2))
            [Turn(WHITE, 0), Turn(WHITE, 1)]

            >>> list(history.turns(start=1, stop=2))
            [Turn(WHITE, 1), Turn(BLACK, 1)]

        :param color: Optional player color indicating which player's turns to return.
        :param start: Optional starting turn number.
        :param stop: Optional stopping turn number.
        :return: An iterable of :class:`Turn` objects that are in the same order as they occurred in the game. If
            `color` is specified, gets the turns only for that player.
        """
        if self.is_empty():
            return

        turn = Turn(color if color is not None else chess.WHITE, start)
        stop_turn = Turn(color if color is not None else chess.WHITE, stop)
        last_turn = self.last_turn()
        while turn <= last_turn and turn < stop_turn:
            if color is None or turn.color == color:
                yield turn
            turn = turn.next

    def is_first_turn(self, turn: Turn):
        """
        Checks whether `turn` is the first turn of the game.

        Examples:
            >>> history.is_first_turn(Turn(BLACK, 0))
            False

            >>> history.is_first_turn(Turn(WHITE, 0))
            True

        :param turn: the :class:`Turn` in question.
        :return: `True` if `turn` is the first turn in the game, `False` otherwise.
        """
        return turn == self.first_turn()

    def first_turn(self, color: Color = None) -> Turn:
        """
        Gets the first turn of the game.

        Examples:
            >>> history.first_turn()
            Turn(WHITE, 0)

            >>> history.first_turn(WHITE)
            Turn(WHITE, 0)

            >>> history.first_turn(BLACK)
            Turn(BLACK, 0)

        :param color: Optional color indicating which player's first turn to return.
        :return: The :class:`Turn` that is the first turn in the game.
        """
        if self.is_empty():
            raise ValueError('GameHistory is empty')

        return Turn(color if color is not None else chess.WHITE, 0)

    def is_last_turn(self, turn: Turn):
        """
        Checks whether `turn` is the last turn of the game.

        Examples:
            >>> history.is_last_turn(Turn(BLACK, 23))
            False

            >>> history.is_last_turn(Turn(WHITE, 24))
            True

        :param turn: the :class:`Turn` in question.
        :return: `True` if `turn` is the last turn in the game, `False` otherwise.
        """
        return turn == self.last_turn()

    def last_turn(self, color: Color = None) -> Turn:
        """
        Gets the last turn of the game.

        Examples:
            >>> history.last_turn()
            Turn(WHITE, 24)

            >>> history.first_turn(WHITE)
            Turn(WHITE, 24)

            >>> history.first_turn(BLACK)
            Turn(BLACK, 23)

        :param color: Optional color indicating which player's last turn to return.
        :return: The :class:`Turn` that is the last turn in the game.
        """
        if self.is_empty():
            raise ValueError('GameHistory is empty')

        if color is not None:
            return Turn(color, len(self._senses[color]) - 1)
        else:
            num_white_turns = len(self._senses[chess.WHITE])
            num_black_turns = len(self._senses[chess.BLACK])
            if num_white_turns > num_black_turns:
                return Turn(chess.WHITE, num_white_turns - 1)
            else:
                return Turn(chess.BLACK, num_black_turns - 1)

    def get_winner_color(self) -> Optional[Color]:
        """
        Returns the color of the player who won the game. If the game is not over, or is over but does not have a winner
         then this will return `None`.

        :return: :class:`Color` of the winner if the game has ended and has a winner, otherwise `None`.
        """
        return self._winner_color

    def get_win_reason(self) -> Optional[WinReason]:
        """
        Returns the reason the player who won won the game. If the game is not over, or is over but does not have a
        winner, then this will return `None`.

        :return: :class:`WinReason` of the winner if the game has ended and has a winner, otherwise `None`.
        """
        return self._win_reason

    def _validate_turn(self, turn: Turn, collection: Mapping[Color, List[T]]):
        if turn.turn_number < 0 or turn.turn_number >= len(collection[turn.color]):
            raise ValueError('{} did not happen in this game.'.format(turn))

    def has_sense(self, turn: Turn) -> bool:
        """
        Checks to see if the game has a sense action for the specified turn. The game may not if it ended because of
        timeout.

        This intended use is to call this before calling :meth:`sense` and :meth:`sense_result`, to verify that there
        is a sense before querying for it.

        Examples:
            >>> history.has_sense(Turn(WHITE, 0))
            True

            >>> history.has_sense(Turn(WHITE, 432))
            False

        :param turn: The :class:`Turn` in question.
        :return: `True` if there is a sense action, `False` otherwise.
        """
        return 0 <= turn.turn_number < len(self._senses[turn.color])

    def sense(self, turn: Turn) -> Optional[Square]:
        """
        Get the sense action on the given turn.

        Examples:
            >>> history.sense(Turn(WHITE, 0))
            E7

            >>> history.sense(Turn(BLACK, 0))
            E2

            >>> history.sense(Turn(WHITE, 1))
            None

        :param turn: The :class:`Turn` in question.
        :return: The executed sense action as a :class:`Square`.
        """
        self._validate_turn(turn, self._senses)
        return self._senses[turn.color][turn.turn_number]

    def sense_result(self, turn: Turn) -> List[Tuple[Square, Optional[chess.Piece]]]:
        """
        Get the result of the sense action on the given turn.

        Examples:
            >>> history.sense(Turn(WHITE, 0))
            B7
            >>> history.sense_result(Turn(WHITE, 0))
            [
                (A8, Piece(ROOK, BLACK)), (B8, Piece(KNIGHT, BLACK)), (C8, Piece(BISHOP, BLACK)),
                (A7, Piece(PAWN, BLACK)), (B7, Piece(PAWN, BLACK)), (C7, Piece(PAWN, BLACK)),
                (A6, None), (B6, None), (C8, None)
            ]
            >>> history.sense(Turn(BLACK, 0))
            None
            >>> history.sense_result(Turn(BLACK, 0))
            []

        :param turn: The :class:`Turn` in question.
        :return: The result of the executed sense action.
        """
        self._validate_turn(turn, self._sense_results)
        return self._sense_results[turn.color][turn.turn_number]

    def has_move(self, turn: Turn) -> bool:
        """
        Checks to see if the game has a move action for the specified turn. The game may not if it ended because of
        timeout.

        This intended use is to call this before calling :meth:`requested_move`, :meth:`taken_move`,
        :meth:`capture_square`, and :meth:`move_result` to verify that there is a move before querying for it.

        Examples:
            >>> history.has_move(Turn(WHITE, 0))
            True

            >>> history.has_move(Turn(WHITE, 432))
            False

        :param turn: The :class:`Turn` in question.
        :return: `True` if there is a move action, `False` otherwise.
        """
        return 0 <= turn.turn_number < len(self._requested_moves[turn.color])

    def requested_move(self, turn: Turn) -> Optional[chess.Move]:
        """
        Get the requested move action on the given turn.

        Examples:
            >>> history.requested_move(Turn(WHITE, 0))
            Move(E2, E4)

            >>> history.requested_move(Turn(BLACK, 0))
            Move(E7, E5)

            >>> history.requested_move(Turn(WHITE, 1))
            None

        :param turn: The :class:`Turn` in question.
        :return: If the player requested to move, then the requested move action as a :class:`chess.Move`, otherwise
            `None` if the player requested to pass.
        """
        self._validate_turn(turn, self._requested_moves)
        return self._requested_moves[turn.color][turn.turn_number]

    def taken_move(self, turn: Turn) -> Optional[chess.Move]:
        """
        Get the executed move action on the given turn. This may be different than the requested move, as the requested
        move may not be legal.

        Examples:
            >>> history.requested_move(Turn(WHITE, 0))
            Move(E2, D3)
            >>> history.taken_move(Turn(WHITE, 0))
            None

            >>> history.requested_move(Turn(WHITE, 1))
            Move(E2, E4)
            >>> history.taken_move(Turn(WHITE, 1))
            Move(E2, E3)

        :param turn: The :class:`Turn` in question.
        :return: `None` if the player requested to pass or made an illegal move, the executed move action as a
            :class:`chess.Move` otherwise.
        """
        self._validate_turn(turn, self._taken_moves)
        return self._taken_moves[turn.color][turn.turn_number]

    def capture_square(self, turn: Turn) -> Optional[Square]:
        """
        Get the square of the opponent's captured piece on the given turn. A capture may not have occurred, in which
        case `None` will be returned.

        Examples:
            >>> history.capture_square(Turn(WHITE, 0))
            None

            >>> history.capture_square(Turn(WHITE, 4))
            E4

        :param turn: The :class:`Turn` in question.
        :return: If a capture occurred, then the :class:`Square` where it occurred, otherwise `None`.
        """
        self._validate_turn(turn, self._capture_squares)
        return self._capture_squares[turn.color][turn.turn_number]

    def move_result(self, turn: Turn) -> Tuple[Optional[chess.Move], Optional[chess.Move], Optional[Square]]:
        """
        Get the full move result in one function. Calls :meth:`requested_move`, :meth:`taken_move`, and
        :meth:`capture_square`.

        Examples:
            >>> history.move_result(Turn(WHITE, 0))
            (Move(E2, E4), Move(E2, E3), None)

        :param turn: The :class:`Turn` in question.
        :return: The requested move, the executed move, and a capture square if there was one.
        """
        return self.requested_move(turn), self.taken_move(turn), self.capture_square(turn)

    def truth_fen_before_move(self, turn: Turn) -> str:
        """
        Get the truth state of the board as a fen string before the move was executed on the given turn. Use
        :meth:`truth_board_before_move` if you want the truth board as a :class:`chess.Board` object.

        Examples:
            >>> history.truth_fen_before_move(Turn(WHITE, 0))
            "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq -"
            >>> history.taken_move(Turn(WHITE, 0))
            Move(E2, E4)
            >>> history.truth_fen_before_move(Turn(BLACK, 0))
            "rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR w KQkq -"

        :param turn: The :class:`Turn` in question.
        :return: The fen of the truth board.
        """
        self._validate_turn(turn, self._fens_before_move)
        return self._fens_before_move[turn.color][turn.turn_number]

    def truth_board_before_move(self, turn: Turn) -> chess.Board:
        """
        Get the truth state of the board as a :class:`chess.Board` before the move was executed on the given turn. Use
        :meth:`truth_fen_before_move` if you want the truth board as a fen string.

        Examples:
            >>> history.truth_board_before_move(Turn(WHITE, 0))
            Board("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq -")
            >>> history.taken_move(Turn(WHITE, 0))
            Move(E2, E4)
            >>> history.truth_fen_before_move(Turn(BLACK, 0))
            Board("rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR w KQkq -")

        :param turn: The :class:`Turn` in question.
        :return: A :class:`chess.Board` object.
        """
        self._validate_turn(turn, self._fens_before_move)
        board = chess.Board(self._fens_before_move[turn.color][turn.turn_number])
        board.turn = turn.color
        return board

    def truth_fen_after_move(self, turn: Turn) -> str:
        """
        Get the truth state of the board as a fen string after the move was executed on the given turn. Use
        :meth:`truth_board_after_move` if you want the truth board as a :class:`chess.Board` object.

        Examples:
            >>> history.truth_fen_before_move(Turn(WHITE, 0))
            "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq -"
            >>> history.taken_move(Turn(WHITE, 0))
            Move(E2, E4)
            >>> history.truth_fen_after_move(Turn(WHITE, 0))
            "rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR w KQkq -"

        :param turn: The :class:`Turn` in question.
        :return: The fen of the truth board.
        """
        self._validate_turn(turn, self._fens_after_move)
        return self._fens_after_move[turn.color][turn.turn_number]

    def truth_board_after_move(self, turn: Turn) -> chess.Board:
        """
        Get the truth state of the board as a :class:`chess.Board` after the move was executed on the given turn. Use
        :meth:`truth_fen_after_move` if you want the truth board as a fen string.

        Examples:
            >>> history.truth_board_before_move(Turn(WHITE, 0))
            Board("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq -")
            >>> history.taken_move(Turn(WHITE, 0))
            Move(E2, E4)
            >>> history.truth_fen_after_move(Turn(WHITE, 0))
            Board("rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR w KQkq -")

        :param turn: The :class:`Turn` in question.
        :return: A :class:`chess.Board` object.
        """
        self._validate_turn(turn, self._fens_after_move)
        board = chess.Board(self._fens_after_move[turn.color][turn.turn_number])
        board.turn = turn.color
        return board

    def collect(self, get_turn_data_fn: Callable[[Turn], T], turns: Iterable[Turn]) -> Iterable[T]:
        """
        Collect data from multiple turns using any of :meth:`sense`, :meth:`sense_result`, :meth:`requested_move`,
        :meth:`taken_move`, :meth:`capture_square`, :meth:`move_result`, :meth:`truth_fen_before_move`,
        :meth:`truth_board_before_move`, :meth:`truth_fen_after_move`, or :meth:`truth_board_after_move`.

        Examples:
            >>> history.collect(history.sense, [Turn(WHITE, 0), Turn(BLACK, 0)])
            [E7, E2]

            >>> history.collect(history.requested_move, history.turns(WHITE))
            [Move(E2, E4), Move(D1, H5), ...]

        :param get_turn_data_fn: One of the getter functions of the history object.
        :param turns: The turns in question.
        :return: A list of the data, where each element is the value of the getter function on the corresponding turn.
        """
        if get_turn_data_fn not in [self.sense, self.sense_result, self.requested_move, self.taken_move,
                                    self.capture_square, self.move_result, self.truth_board_before_move,
                                    self.truth_board_after_move, self.truth_fen_before_move, self.truth_fen_after_move]:
            raise ValueError('get_turn_data_fn must be one of the history getter functions')
        for turn in turns:
            yield get_turn_data_fn(turn)

    def __eq__(self, other):
        if not isinstance(other, GameHistory):
            return NotImplemented

        senses_equal = self._senses == other._senses and self._sense_results == other._sense_results
        moves_equal = self._requested_moves == other._requested_moves and self._taken_moves == other._taken_moves and self._capture_squares == other._capture_squares
        fens_equal = self._fens_before_move == other._fens_before_move and self._fens_after_move == other._fens_after_move
        results_equal = self._win_reason == other._win_reason and self._winner_color == other._winner_color

        return senses_equal and moves_equal and fens_equal and results_equal


class GameHistoryEncoder(ChessJSONEncoder):
    def default(self, o):
        if isinstance(o, GameHistory):
            return {
                'type': 'GameHistory',
                'white_name': o._white_name,
                'black_name': o._black_name,
                'senses': o._senses,
                'sense_results': o._sense_results,
                'requested_moves': o._requested_moves,
                'taken_moves': o._taken_moves,
                'capture_squares': o._capture_squares,
                'fens_before_move': o._fens_before_move,
                'fens_after_move': o._fens_after_move,
                'winner_color': o._winner_color,
                'win_reason': o._win_reason,
            }
        return super().default(o)


class GameHistoryDecoder(ChessJSONDecoder):
    def _object_hook(self, obj):
        if 'type' in obj and obj['type'] == 'GameHistory':
            for key in ['senses', 'sense_results', 'requested_moves', 'taken_moves', 'capture_squares',
                        'fens_before_move', 'fens_after_move']:
                obj[key] = {True: obj[key]['true'], False: obj[key]['false']}
            history = GameHistory()
            history._white_name = obj['white_name']
            history._black_name = obj['black_name']
            history._senses = obj['senses']
            history._sense_results = obj['sense_results']
            history._requested_moves = obj['requested_moves']
            history._taken_moves = obj['taken_moves']
            history._capture_squares = obj['capture_squares']
            history._fens_before_move = obj['fens_before_move']
            history._fens_after_move = obj['fens_after_move']
            history._winner_color = obj['winner_color']
            history._win_reason = obj['win_reason']

            for color, sense_results in history._sense_results.items():
                for result in sense_results:
                    for i in range(len(result)):
                        result[i] = tuple(result[i])
            return history

        return super()._object_hook(obj)



import argparse
import datetime
import traceback
import chess
from reconchess import load_player, play_local_game, LocalGame


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('white_bot_path', help='path to white bot source file')
    parser.add_argument('black_bot_path', help='path to black bot source file')
    parser.add_argument('--seconds_per_player', default=900, type=float,
                        help='number of seconds each player has to play the entire game.')
    args = parser.parse_args()

    white_bot_name, white_player_cls = load_player(args.white_bot_path)
    black_bot_name, black_player_cls = load_player(args.black_bot_path)

    game = LocalGame(args.seconds_per_player)

    try:
        winner_color, win_reason, history = play_local_game(white_player_cls(), black_player_cls(), game=game)

        winner = 'Draw' if winner_color is None else chess.COLOR_NAMES[winner_color]
    except:
        traceback.print_exc()
        game.end()

        winner = 'ERROR'
        history = game.get_game_history()

    print('Game Over!')
    print('Winner: {}!'.format(winner))

    timestamp = datetime.datetime.now().strftime('%Y_%m_%d-%H_%M_%S')

    replay_path = '{}-{}-{}-{}.json'.format(white_bot_name, black_bot_name, winner, timestamp)
    print('Saving replay to {}...'.format(replay_path))
    history.save(replay_path)


if __name__ == '__main__':
    main()



__version__ = '1.6.9'

from .game import Game, LocalGame, RemoteGame
from .player import Player, load_player
from .types import *
from .utilities import is_illegal_castle, is_psuedo_legal_castle, ChessJSONEncoder, ChessJSONDecoder
from .play import play_local_game, play_remote_game, play_turn, notify_opponent_move_results, play_sense, play_move
from .history import Turn, GameHistory, GameHistoryEncoder, GameHistoryDecoder
import chess



try:
    # requests uses simplejson if its present, so we need to do the same.
    # see https://github.com/psf/requests/issues/4842
    import simplejson as json
except ImportError:
    import json
import chess
from .types import *

BACK_RANKS = list(chess.SquareSet(chess.BB_BACKRANKS))


def add_pawn_queen_promotion(board: chess.Board, move: chess.Move) -> chess.Move:
    piece = board.piece_at(move.from_square)
    if piece is not None and piece.piece_type == chess.PAWN and move.to_square in BACK_RANKS and move.promotion is None:
        move = chess.Move(move.from_square, move.to_square, chess.QUEEN)
    return move


def is_psuedo_legal_castle(board: chess.Board, move: chess.Move):
    return board.is_castling(move) and not is_illegal_castle(board, move)


def is_illegal_castle(board: chess.Board, move: chess.Move) -> bool:
    if not board.is_castling(move):
        return False

    # illegal without kingside rights
    if board.is_kingside_castling(move) and not board.has_kingside_castling_rights(board.turn):
        return True

    # illegal without queenside rights
    if board.is_queenside_castling(move) and not board.has_queenside_castling_rights(board.turn):
        return True

    # illegal if any pieces are between king & rook
    rook_square = chess.square(7 if board.is_kingside_castling(move) else 0, chess.square_rank(move.from_square))
    between_squares = chess.SquareSet(chess.between(move.from_square, rook_square))
    if any(map(lambda s: board.piece_at(s), between_squares)):
        return True

    # its legal
    return False


def slide_move(board: chess.Board, move: chess.Move) -> Optional[chess.Move]:
    # We iterate longest to shortest so the revised move is the longest pseudo-legal move.
    # If the to-square < from-square then we want our list in sorted order.
    # Otherwise we want it in reverse order.
    # In either case, we need to add the to-square to the front of our list manually.
    squares = chess.SquareSet(chess.between(move.from_square, move.to_square))
    if move.to_square > move.from_square:
        squares = reversed(squares)
    for slide_square in [move.to_square] + list(squares):
        revised = chess.Move(move.from_square, slide_square, move.promotion)
        if board.is_pseudo_legal(revised):
            return revised
    return None


def capture_square_of_move(board: chess.Board, move: Optional[chess.Move]) -> Optional[Square]:
    capture_square = None
    if move is not None and board.is_capture(move):
        if board.is_en_passant(move):
            # taken from :func:`chess.Board.push()`
            down = -8 if board.turn == chess.WHITE else 8
            capture_square = board.ep_square + down
        else:
            capture_square = move.to_square
    return capture_square


def without_opponent_pieces(board: chess.Board) -> chess.Board:
    """Returns a copy of `board` with the opponent's pieces removed."""
    mine = board.occupied_co[board.turn]
    return board.transform(lambda bb: bb & mine)


def moves_without_opponent_pieces(board: chess.Board) -> List[chess.Move]:
    """Generates moves on `board` with the opponent's pieces removed."""
    return list(without_opponent_pieces(board).generate_pseudo_legal_moves())


def pawn_capture_moves_on(board: chess.Board) -> List[chess.Move]:
    """Generates all pawn captures on `board`, even if there is no piece to capture. All promotion moves are included."""
    pawn_capture_moves = []

    for pawn_square in board.pieces(chess.PAWN, board.turn):
        for attacked_square in board.attacks(pawn_square):
            # skip this square if one of our own pieces are on the square
            attacked_piece = board.piece_at(attacked_square)
            if attacked_piece and attacked_piece.color == board.turn:
                continue

            pawn_capture_moves.append(chess.Move(pawn_square, attacked_square))

            # add in promotion moves
            if attacked_square in BACK_RANKS:
                for piece_type in chess.PIECE_TYPES[1:-1]:
                    pawn_capture_moves.append(chess.Move(pawn_square, attacked_square, promotion=piece_type))

    return pawn_capture_moves


def revise_move(board, move):
    # if its a legal move, don't change it at all. note that board.generate_psuedo_legal_moves() does not
    # include psuedo legal castles
    if board.is_pseudo_legal(move) or is_psuedo_legal_castle(board, move):
        return move

    # note: if there are pieces in the way, we DONT capture them
    if is_illegal_castle(board, move):
        return None

    # if the piece is a sliding piece, slide it as far as it can go
    piece = board.piece_at(move.from_square)
    if piece.piece_type in [chess.PAWN, chess.ROOK, chess.BISHOP, chess.QUEEN]:
        move = slide_move(board, move)

    return move if board.is_pseudo_legal(move) else None


def move_actions(board: chess.Board) -> List[chess.Move]:
    """
    :return: List of moves that are possible with only knowledge of your pieces
    """
    return moves_without_opponent_pieces(board) + pawn_capture_moves_on(board)


class ChessJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, chess.Piece):
            return {
                'type': 'Piece',
                'value': o.symbol(),
            }
        elif isinstance(o, chess.Move):
            return {
                'type': 'Move',
                'value': o.uci(),
            }
        elif isinstance(o, chess.Board):
            return {
                'type': 'Board',
                'value': o.fen(),
            }
        elif isinstance(o, WinReason):
            return {
                'type': 'WinReason',
                'value': o.name,
            }
        return super().default(o)


class ChessJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        hook = self._object_hook
        if 'object_hook' in kwargs:
            original_hook = kwargs.pop('object_hook')
            hook = lambda obj: self._object_hook(original_hook(obj))
        super().__init__(object_hook=hook, *args, **kwargs)

    def _object_hook(self, obj):
        if 'type' in obj:
            if obj['type'] == 'Piece':
                return chess.Piece.from_symbol(obj['value'])
            elif obj['type'] == 'Move':
                return chess.Move.from_uci(obj['value'])
            elif obj['type'] == 'Board':
                return chess.Board(fen=obj['value'])
            elif obj['type'] == 'WinReason':
                return WinReason[obj['value']]
        return obj



import argparse
import chess
import datetime
import random
from reconchess import play_remote_game
from reconchess.scripts.rc_connect import RBCServer, ask_for_auth
from reconchess.scripts.rc_play import UIPlayer


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--color', default='random', choices=['white', 'black', 'random'],
                        help='The color you want to play as.')
    parser.add_argument('--server-url', default='https://rbc.jhuapl.edu', help='URL of the server.')
    args = parser.parse_args()

    auth = ask_for_auth()

    server = RBCServer(args.server_url, auth)

    usernames = server.get_active_users()

    if auth[0] in usernames:
        usernames.remove(auth[0])

    if len(usernames) == 0:
        print('No active users.')
        quit()

    for i, username in enumerate(usernames):
        print('[{}] {}'.format(i, username))
    i = int(input('Choose opponent: '))

    color = chess.WHITE
    if args.color == 'black' or (args.color == 'random' and random.uniform(0, 1) < 0.5):
        color = chess.BLACK

    game_id = server.send_invitation(usernames[i], color)

    winner_color, win_reason, game_history = play_remote_game(args.server_url, game_id, auth, UIPlayer())

    winner = 'Draw' if winner_color is None else chess.COLOR_NAMES[winner_color]
    timestamp = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
    game_history.save('{}_{}_{}.json'.format(timestamp, game_id, winner))


if __name__ == '__main__':
    main()



import os
import sys
import importlib
import inspect
from abc import abstractmethod
import chess
from .types import *
from .history import GameHistory


class Player(object):
    """
    Base class of a player of Recon Chess. Implementation of a player is done by sub classing this class, and
    implementing each of the methods detailed below. For examples see `examples`.

    The order in which each of the methods are called looks roughly like this:

    #. :meth:`handle_game_start()`
    #. :meth:`handle_opponent_move_result()`
    #. :meth:`choose_sense()`
    #. :meth:`handle_sense_result()`
    #. :meth:`choose_move()`
    #. :meth:`handle_move_result()`
    #. :meth:`handle_game_end()`

    Note that the :meth:`handle_game_start()` and :meth:`handle_game_end()` methods are only called at the start and
    the end of the game respectively. The rest are called repeatedly for each of your turns.
    """

    @abstractmethod
    def handle_game_start(self, color: Color, board: chess.Board, opponent_name: str):
        """
        Provides a place to initialize game wide structures like boards, and initialize data that depends on what
        color you are playing as.

        Called when the game starts.

        The board is provided to allow starting a game from different board positions.

        :param color: The color that you are playing as. Either :data:`chess.WHITE` or :data:`chess.BLACK`.
        :param board: The initial board of the game. See :class:`chess.Board`.
        :param opponent_name: The name of your opponent.
        """
        pass

    @abstractmethod
    def handle_opponent_move_result(self, captured_my_piece: bool, capture_square: Optional[Square]):
        """
        Provides information about what happened on the opponents turn.

        Called at the start of your turn.

        Example implementation: ::

            def handle_opponent_move_result(self, captured_my_piece: bool, capture_square: Optional[Square]):
                if captured_my_piece:
                    self.board.remove_piece_at(capture_square)

        :param captured_my_piece: If the opponent captured one of your pieces, then `True`, otherwise `False`.
        :param capture_square: If a capture occurred, then the :class:`Square` your piece was captured on,
            otherwise `None`.
        """
        pass

    @abstractmethod
    def choose_sense(self, sense_actions: List[Square], move_actions: List[chess.Move], seconds_left: float) -> \
            Optional[Square]:
        """
        The method to implement your sensing strategy. The chosen sensing action should be returned from this function.
        I.e. the value returned is the square at the center of the 3x3 sensing area you want to sense. The returned
        square must be one of the squares in the `sense_actions` parameter.

        You can pass instead of sensing by returning `None` from this function.

        Move actions are provided through `move_actions` in case you want to sense based on a potential move.

        Called after :meth:`handle_opponent_move_result()`.

        Example implementation: ::

            def choose_sense(self, sense_actions: List[Square], move_actions: List[chess.Move], seconds_left: float) -> Square:
                return random.choice(sense_actions)

        :param sense_actions: A :class:`list` containing the valid squares to sense over.
        :param move_actions: A :class:`list` containing the valid moves that can be returned in :meth:`choose_move()`.
        :param seconds_left: The time in seconds you have left to use in the game.
        :return: a :class:`Square` that is the center of the 3x3 sensing area you want to get information about.
        """
        pass

    @abstractmethod
    def handle_sense_result(self, sense_result: List[Tuple[Square, Optional[chess.Piece]]]):
        """
        Provides the result of the sensing action. Each element in `sense_result` is a square and the corresponding
        :class:`chess.Piece` found on that square. If there is no piece on the square, then the piece will be `None`.

        Called after :meth:`choose_sense()`.

        Example implementation: ::

            def handle_sense_result(self, sense_result: List[Tuple[Square, Optional[chess.Piece]]]):
                for square, piece in sense_result:
                    if piece is None:
                        self.board.remove_piece_at(square)
                    else:
                        self.board.set_piece_at(square, piece)

        :param sense_result: The result of the sense. A `list` of :class:`Square` and an optional :class:`chess.Piece`.
        """
        pass

    @abstractmethod
    def choose_move(self, move_actions: List[chess.Move], seconds_left: float) -> Optional[chess.Move]:
        """
        The method to implement your movement strategy. The chosen movement action should be returned from this function.
        I.e. the value returned is the move to make. The returned move must be one of the moves in the `move_actions`
        parameter.

        The pass move is legal, and is executed by returning `None` from this method.

        Called after :meth:`handle_sense_result()`.

        Example implementation: ::

            def choose_move(self, move_actions: List[chess.Move], seconds_left: float) -> Optional[chess.Move]:
                return random.choice(move_actions)

        :param move_actions: A `list` containing the valid :class:`chess.Move` you can choose.
        :param seconds_left: The time in seconds you have left to use in the game.
        :return: The :class:`chess.Move` to make.
        """
        pass

    @abstractmethod
    def handle_move_result(self, requested_move: Optional[chess.Move], taken_move: Optional[chess.Move],
                           captured_opponent_piece: bool, capture_square: Optional[Square]):
        """
        Provides the result of the movement action. The `requested_move` is the move returned from :meth:`choose_move()`,
        and is provided for ease of use. `taken_move` is the move that was actually performed. Note that `taken_move`,
        can be different from `requested_move`, due to the uncertainty aspect.

        Called after :meth:`choose_move()`.

        Example implementation: ::

                def handle_move_result(self, requested_move: chess.Move, taken_move: chess.Move,
                           captured_opponent_piece: bool, capture_square: Optional[Square]):
                    if taken_move is not None:
                        self.board.push(taken_move)

        Note: In the case of playing games on a server, this method is invoked during your opponents turn. This means
        in most cases this method will not use your play time. However if the opponent finishes their turn before
        this method completes, then time will be counted against you.

        :param requested_move: The :class:`chess.Move` you requested in :meth:`choose_move()`.
        :param taken_move: The :class:`chess.Move` that was actually applied by the game if it was a valid move,
            otherwise `None`.
        :param captured_opponent_piece: If `taken_move` resulted in a capture, then `True`, otherwise `False`.
        :param capture_square: If a capture occurred, then the :class:`Square` that the opponent piece was taken on,
            otherwise `None`.
        """
        pass

    @abstractmethod
    def handle_game_end(self, winner_color: Optional[Color], win_reason: Optional[WinReason],
                        game_history: GameHistory):
        """
        Provides the results of the game when it ends. You can use this for post processing the results of the game.

        :param winner_color: If the game was a draw, then `None`, otherwise, the color of the player who won the game.
        :param win_reason: If the game was a draw, then `None`, otherwise the reason the game ended specified as
            :class:`WinReason`
        :param game_history: :class:`GameHistory` object for the game, from which you can get the actions
            each side has taken over the course of the game.
        """
        pass


def load_player(source_path: str) -> Tuple[str, Type[Player]]:
    """
    Loads a subclass of the Player class that is contained in a python source file or python module.
    There should only be 1 such subclass in the file or module. If there are more than 1 subclasses, then you have
    to define a function named `get_player` in the same module that returns the subclass to use.

    Example of single class definition: ::

        # this will import fine
        class MyBot(Player):
            ...

    Example of multiple class definition: ::

        class MyBot1(Player):
            ...

        class MyBot2(Player):
            ...

        # need to define this function!
        def get_player():
            return MyBot1

    Example of another situation where you may need to define `get_player`: ::

        from my_helper_module import MyPlayerBaseClass

        class MyBot1(MyPlayerBaseClass):
            ...

        # you need to define this because both MyBot1 and MyPlayerBaseClass are subclasses of Player
        def get_player():
            return MyBot1

    Example usage: ::

        name, cls = load_player('my_player.py')
        player = cls()

        name, cls = load_player('reconchess.bots.random_bot')
        player = cls()

    :param source_path: the path to the source file to load
    :return: Tuple where the first element is the name of the loaded class, and the second element is the class type
    """
    if os.path.exists(source_path):
        # get the path to the main source file
        abs_source_path = os.path.abspath(source_path)

        # insert the directory of the bot source file into system path so we can import it
        # note: insert it first so we know we are searching this first
        sys.path.insert(0, os.path.dirname(abs_source_path))

        # import_module expects a module name, so remove the extension
        module_name = os.path.splitext(os.path.basename(abs_source_path))[0]
    else:
        module_name = source_path

    module = importlib.import_module(module_name)
    players = inspect.getmembers(module, lambda o: inspect.isclass(o) and issubclass(o, Player) and o != Player)
    get_player_fns = inspect.getmembers(module, lambda o: inspect.isfunction(o) and o.__name__ == 'get_player')
    if len(players) == 0:
        raise RuntimeError('{} did not contain any subclasses of {}'.format(source_path, Player))
    elif len(players) > 1 and len(get_player_fns) != 1:
        msg = '{} contained multiple subclasses of {}: {}'.format(source_path, Player, players)
        msg += ', but no get_player function was defined. See documentation for reconchess.load_player'
        raise RuntimeError(msg)

    if len(players) == 1:
        return players[0]
    else:
        _, get_player_fn = get_player_fns[0]
        cls = get_player_fn()
        return cls.__name__, cls



import chess.engine
import random
from reconchess import *
import os

STOCKFISH_ENV_VAR = 'STOCKFISH_EXECUTABLE'


class TroutBot(Player):
    """
    TroutBot uses the Stockfish chess engine to choose moves. In order to run TroutBot you'll need to download
    Stockfish from https://stockfishchess.org/download/ and create an environment variable called STOCKFISH_EXECUTABLE
    that is the path to the downloaded Stockfish executable.
    """

    def __init__(self):
        self.board = None
        self.color = None
        self.my_piece_captured_square = None

        # make sure stockfish environment variable exists
        if STOCKFISH_ENV_VAR not in os.environ:
            raise KeyError(
                'TroutBot requires an environment variable called "{}" pointing to the Stockfish executable'.format(
                    STOCKFISH_ENV_VAR))

        # make sure there is actually a file
        stockfish_path = os.environ[STOCKFISH_ENV_VAR]
        if not os.path.exists(stockfish_path):
            raise ValueError('No stockfish executable found at "{}"'.format(stockfish_path))

        # initialize the stockfish engine
        self.engine = chess.engine.SimpleEngine.popen_uci(stockfish_path, setpgrp=True)

    def handle_game_start(self, color: Color, board: chess.Board, opponent_name: str):
        self.board = board
        self.color = color

    def handle_opponent_move_result(self, captured_my_piece: bool, capture_square: Optional[Square]):
        # if the opponent captured our piece, remove it from our board.
        self.my_piece_captured_square = capture_square
        if captured_my_piece:
            self.board.remove_piece_at(capture_square)

    def choose_sense(self, sense_actions: List[Square], move_actions: List[chess.Move], seconds_left: float) -> \
            Optional[Square]:
        # if our piece was just captured, sense where it was captured
        if self.my_piece_captured_square:
            return self.my_piece_captured_square

        # if we might capture a piece when we move, sense where the capture will occur
        future_move = self.choose_move(move_actions, seconds_left)
        if future_move is not None and self.board.piece_at(future_move.to_square) is not None:
            return future_move.to_square

        # otherwise, just randomly choose a sense action, but don't sense on a square where our pieces are located
        for square, piece in self.board.piece_map().items():
            if piece.color == self.color:
                sense_actions.remove(square)
        return random.choice(sense_actions)

    def handle_sense_result(self, sense_result: List[Tuple[Square, Optional[chess.Piece]]]):
        # add the pieces in the sense result to our board
        for square, piece in sense_result:
            self.board.set_piece_at(square, piece)

    def choose_move(self, move_actions: List[chess.Move], seconds_left: float) -> Optional[chess.Move]:
        # if we might be able to take the king, try to
        enemy_king_square = self.board.king(not self.color)
        if enemy_king_square:
            # if there are any ally pieces that can take king, execute one of those moves
            enemy_king_attackers = self.board.attackers(self.color, enemy_king_square)
            if enemy_king_attackers:
                attacker_square = enemy_king_attackers.pop()
                return chess.Move(attacker_square, enemy_king_square)

        # otherwise, try to move with the stockfish chess engine
        try:
            self.board.turn = self.color
            self.board.clear_stack()
            result = self.engine.play(self.board, chess.engine.Limit(time=0.5))
            return result.move
        except chess.engine.EngineTerminatedError:
            print('Stockfish Engine died')
        except chess.engine.EngineError:
            print('Stockfish Engine bad state at "{}"'.format(self.board.fen()))

        # if all else fails, pass
        return None

    def handle_move_result(self, requested_move: Optional[chess.Move], taken_move: Optional[chess.Move],
                           captured_opponent_piece: bool, capture_square: Optional[Square]):
        # if a move was executed, apply it to our board
        if taken_move is not None:
            self.board.push(taken_move)

    def handle_game_end(self, winner_color: Optional[Color], win_reason: Optional[WinReason],
                        game_history: GameHistory):
        try:
            # if the engine is already terminated then this call will throw an exception
            self.engine.quit()
        except chess.engine.EngineTerminatedError:
            pass



import random
from reconchess import *

# move sequences from white's perspective, flipped at runtime if playing as black
QUICK_ATTACKS = [
    # queen-side knight attacks
    [chess.Move(chess.B1, chess.C3), chess.Move(chess.C3, chess.B5), chess.Move(chess.B5, chess.D6),
     chess.Move(chess.D6, chess.E8)],
    [chess.Move(chess.B1, chess.C3), chess.Move(chess.C3, chess.E4), chess.Move(chess.E4, chess.F6),
     chess.Move(chess.F6, chess.E8)],

    # king-side knight attacks
    [chess.Move(chess.G1, chess.H3), chess.Move(chess.H3, chess.F4), chess.Move(chess.F4, chess.H5),
     chess.Move(chess.H5, chess.F6), chess.Move(chess.F6, chess.E8)],

    # four move mates
    [chess.Move(chess.E2, chess.E4), chess.Move(chess.F1, chess.C4), chess.Move(chess.D1, chess.H5), chess.Move(
        chess.C4, chess.F7), chess.Move(chess.F7, chess.E8), chess.Move(chess.H5, chess.E8)],
]


def flipped_move(move):
    def flipped(square):
        return chess.square(chess.square_file(square), 7 - chess.square_rank(square))

    return chess.Move(from_square=flipped(move.from_square), to_square=flipped(move.to_square),
                      promotion=move.promotion, drop=move.drop)


class AttackerBot(Player):
    def __init__(self):
        self.move_sequence = random.choice(QUICK_ATTACKS)

    def handle_game_start(self, color: Color, board: chess.Board, opponent_name: str):
        if color == chess.BLACK:
            self.move_sequence = list(map(flipped_move, self.move_sequence))

    def handle_opponent_move_result(self, captured_my_piece: bool, capture_square: Optional[Square]):
        pass

    def choose_sense(self, sense_actions: List[Square], move_actions: List[chess.Move], seconds_left: float) -> \
            Optional[Square]:
        return random.choice(sense_actions)

    def handle_sense_result(self, sense_result: List[Tuple[Square, Optional[chess.Piece]]]):
        pass

    def choose_move(self, move_actions: List[chess.Move], seconds_left: float) -> Optional[chess.Move]:
        while len(self.move_sequence) > 0 and self.move_sequence[0] not in move_actions:
            self.move_sequence.pop(0)

        if len(self.move_sequence) == 0:
            # pass... we failed so give up
            return None
        else:
            return self.move_sequence.pop(0)

    def handle_move_result(self, requested_move: Optional[chess.Move], taken_move: Optional[chess.Move],
                           captured_opponent_piece: bool, capture_square: Optional[Square]):
        pass

    def handle_game_end(self, winner_color: Optional[Color], win_reason: Optional[WinReason],
                        game_history: GameHistory):
        pass



import pkg_resources
import traceback
import argparse
import random
import contextlib
from reconchess import *
import datetime

# block output from pygame
with contextlib.redirect_stdout(None):
    import pygame


class Window:
    LIGHT_COLOR = (240, 217, 181)
    DARK_COLOR = (181, 136, 99)

    HIGHLIGHT_COLOR = (255, 255, 0)
    CAPTURE_COLOR = (255, 0, 0)

    LIGHT_SQUARES = list(chess.SquareSet(chess.BB_LIGHT_SQUARES))
    DARK_SQUARES = list(chess.SquareSet(chess.BB_DARK_SQUARES))

    def __init__(self):
        self.fps = 30
        self.width = 640
        self.height = 640
        self.square_size = self.width // 8

        self.image_by_piece = {}
        for color in chess.COLORS:
            for piece_type in chess.PIECE_TYPES:
                piece = chess.Piece(piece_type, color)

                img_path = 'res/{}/{}.png'.format(chess.COLOR_NAMES[color], piece.symbol())
                full_path = pkg_resources.resource_filename('reconchess', img_path)

                img = pygame.image.load(full_path)
                img = pygame.transform.scale(img, (self.square_size, self.square_size))
                self.image_by_piece[piece] = img

        pygame.init()
        pygame.display.set_caption('Recon Chess')
        pygame.display.set_icon(
            pygame.transform.scale(self.image_by_piece[chess.Piece(chess.KING, chess.WHITE)], (32, 32)))

        self.clock = pygame.time.Clock()
        self.screen = pygame.display.set_mode((self.width, self.height))
        self.background = pygame.Surface((self.screen.get_size()))

        self.perspective = chess.WHITE

        self.callback_by_event = {}

    def register_callback(self, event, callback):
        self.callback_by_event[event] = callback

    def coords_to_square(self, x, y):
        file = (x // self.square_size)
        if self.perspective == chess.WHITE:
            rank = ((self.height - y) // self.square_size)
        else:
            rank = (y // self.square_size)
        return chess.square(file, rank)

    def square_to_coords(self, square):
        rank = chess.square_rank(square)
        file = chess.square_file(square)
        x = file * self.square_size
        if self.perspective == chess.WHITE:
            y = (7 - rank) * self.square_size
        else:
            y = rank * self.square_size
        return x, y

    def square_rect(self, square):
        x, y = self.square_to_coords(square)
        return x, y, self.square_size, self.square_size

    def mouse_on_board(self):
        return pygame.mouse.get_focused()

    def right_pressed(self):
        return pygame.mouse.get_pressed()[2]

    def left_pressed(self):
        return pygame.mouse.get_pressed()[0]

    def draw(self, board, highlighted_squares=None, capture_squares=None, ignored_square=None, floating_piece=None):
        self.background.fill((0, 0, 0))

        self.draw_board()
        self.draw_pieces_on(board, ignored_square)

        if highlighted_squares is not None:
            for square in highlighted_squares:
                self.draw_highlight(square)

        if capture_squares is not None:
            for square in capture_squares:
                if square is not None:
                    self.draw_capture(square)

        if floating_piece:
            self.draw_piece_at(floating_piece, *pygame.mouse.get_pos())

        self.screen.blit(self.background, (0, 0))
        pygame.display.flip()
        self.clock.tick(self.fps)
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                quit()
            if event.type in self.callback_by_event:
                self.callback_by_event[event.type](event)

    def draw_board(self):
        for square in chess.SQUARES:
            color = self.LIGHT_COLOR if square in self.LIGHT_SQUARES else self.DARK_COLOR
            pygame.draw.rect(self.background, color, self.square_rect(square))

    def draw_pieces_on(self, board, ignored_square=None):
        for square in chess.SQUARES:
            if square == ignored_square:
                continue

            piece = board.piece_at(square)
            if piece is not None:
                image = self.image_by_piece[piece]
                self.background.blit(image, self.square_rect(square))

    def draw_piece_at(self, piece, x, y):
        x -= self.square_size / 2
        y -= self.square_size / 2
        if piece is not None:
            image = self.image_by_piece[piece]
            self.background.blit(image, (x, y, self.square_size, self.square_size))

    def draw_highlight(self, square):
        pygame.draw.rect(self.background, self.HIGHLIGHT_COLOR, self.square_rect(square), 3)

    def draw_capture(self, square):
        pygame.draw.rect(self.background, self.CAPTURE_COLOR, self.square_rect(square), 3)


class UIPlayer(Player):
    def __init__(self):
        self.window = Window()
        self.window.register_callback(pygame.MOUSEBUTTONUP, self._handle_mouse_up)

        self.board = None
        self.color = None
        self.ally_capture_square = None
        self.enemy_capture_square = None

    def handle_game_start(self, color: Color, board: chess.Board, opponent_name: str):
        self.board = board
        self.color = color
        self.window.perspective = color

    def handle_opponent_move_result(self, captured_my_piece: bool, capture_square: Optional[Square]):
        self.ally_capture_square = capture_square
        self.board.turn = self.color

        if captured_my_piece:
            self.board.remove_piece_at(capture_square)

        self.window.draw(self.board, capture_squares=[self.enemy_capture_square, self.ally_capture_square])

    def choose_sense(self, sense_actions: List[Square], move_actions: List[chess.Move], seconds_left: float) -> Square:
        while True:
            if self.window.mouse_on_board():
                square = self.window.coords_to_square(*pygame.mouse.get_pos())
                if square in sense_actions:
                    sense_area = self._squares_in_sense_around(square)
                else:
                    sense_area = []
            else:
                sense_area = []

            self.window.draw(self.board, highlighted_squares=sense_area,
                             capture_squares=[self.enemy_capture_square, self.ally_capture_square])

            if self.window.left_pressed() and self.window.mouse_on_board():
                square = self.window.coords_to_square(*pygame.mouse.get_pos())
                if square in sense_actions:
                    return square

    def _squares_in_sense_around(self, center):
        rank, file = chess.square_rank(center), chess.square_file(center)
        squares = []
        for dr in [-1, 0, 1]:
            for df in [-1, 0, 1]:
                r, f = rank + dr, file + df
                if 0 <= r < 8 and 0 <= f < 8:
                    squares.append(chess.square(f, r))
        return squares

    def handle_sense_result(self, sense_result: List[Tuple[Square, Optional[chess.Piece]]]):
        for square, piece in sense_result:
            self.board.set_piece_at(square, piece)

        self.window.draw(self.board, capture_squares=[self.enemy_capture_square, self.ally_capture_square])

    def choose_move(self, move_actions: List[chess.Move], seconds_left: float) -> Optional[chess.Move]:
        selected_square = None
        floating_piece = None

        while True:
            if selected_square is not None:
                to_squares = [move.to_square for move in move_actions if move.from_square == selected_square]
            elif self.window.mouse_on_board():
                square = self.window.coords_to_square(*pygame.mouse.get_pos())
                to_squares = [move.to_square for move in move_actions if move.from_square == square]
            else:
                to_squares = []

            self.window.draw(self.board, highlighted_squares=to_squares, ignored_square=selected_square,
                             floating_piece=floating_piece,
                             capture_squares=[self.enemy_capture_square, self.ally_capture_square])

            # picking up piece
            if self.window.left_pressed() and self.window.mouse_on_board():
                mouse_square = self.window.coords_to_square(*pygame.mouse.get_pos())
                piece = self.board.piece_at(mouse_square)
                if selected_square is None and piece is not None and piece.color == self.color:
                    selected_square = mouse_square
                    floating_piece = self.board.piece_at(mouse_square)

            # dropping/playing piece
            if selected_square is not None and not self.window.left_pressed():
                if not self.window.mouse_on_board():
                    selected_square = None
                    floating_piece = None
                else:
                    to_square = self.window.coords_to_square(*pygame.mouse.get_pos())
                    move = chess.Move(selected_square, to_square)
                    promotion_move = chess.Move(selected_square, to_square, promotion=chess.QUEEN)
                    if move in move_actions:
                        return move
                    elif promotion_move in move_actions:
                        return promotion_move
                    else:
                        selected_square = None
                        floating_piece = None

    def handle_move_result(self, requested_move: Optional[chess.Move], taken_move: Optional[chess.Move],
                           captured_opponent_piece: bool, capture_square: Optional[Square]):
        if taken_move is not None:
            self.board.push(taken_move)

        self.enemy_capture_square = capture_square

        self.window.draw(self.board, capture_squares=[self.enemy_capture_square, self.ally_capture_square])

    def handle_game_end(self, winner_color: Optional[Color], win_reason: Optional[WinReason],
                        game_history: GameHistory):
        pass

    def _handle_mouse_up(self, event):
        if event.button == 3 and self.window.mouse_on_board():
            square = self.window.coords_to_square(*pygame.mouse.get_pos())
            self._rotate_piece_at(square)

    def _rotate_piece_at(self, square):
        piece = self.board.piece_at(square)
        if piece is not None and piece.color == self.color:
            return

        if piece is None:
            self.board.set_piece_at(square, chess.Piece(chess.PAWN, not self.color))
        else:
            if piece.piece_type == chess.KING:
                self.board.set_piece_at(square, None)
            else:
                self.board.set_piece_at(square, chess.Piece(piece.piece_type + 1, not self.color))


def main():
    parser = argparse.ArgumentParser(description='Allows you to play against a bot. Useful for testing and debugging.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bot_path', help='Path to bot source file.')
    parser.add_argument('--color', default='random', choices=['white', 'black', 'random'],
                        help='The color you want to play as.')
    parser.add_argument('--seconds_per_player', default=900, type=float,
                        help='number of seconds each player has to play the entire game.')
    args = parser.parse_args()

    bot_name, bot_constructor = load_player(args.bot_path)
    bot = bot_constructor()
    player = UIPlayer()

    players = [player, bot]
    player_names = ['Human', bot_name]
    if args.color == 'black' or (args.color == 'random' and random.uniform(0, 1) < 0.5):
        players.reverse()
        player_names.reverse()

    game = LocalGame(args.seconds_per_player)

    try:
        winner_color, win_reason, history = play_local_game(players[0], players[1], game)
        winner = 'Draw' if winner_color is None else chess.COLOR_NAMES[winner_color]
    except:
        traceback.print_exc()
        game.end()

        winner = 'ERROR'
        history = game.get_game_history()

    print('Game Over!')
    print('Winner: {}!'.format(winner))

    timestamp = datetime.datetime.now().strftime('%Y_%m_%d-%H_%M_%S')

    replay_path = '{}-{}-{}-{}.json'.format(player_names[0], player_names[1], winner, timestamp)
    print('Saving replay to {}...'.format(replay_path))
    history.save(replay_path)


if __name__ == '__main__':
    main()



import argparse
import chess
from reconchess import load_player, LocalGame, GameHistory, Player, Color, play_local_game


def playback(game_history: GameHistory, player: Player, color: Color):
    game = LocalGame()

    opponent_name = game_history.get_white_player_name()
    if color == chess.WHITE:
        opponent_name = game_history.get_black_player_name()
    player.handle_game_start(color, game.board.copy(), opponent_name)
    game.start()

    turn = game_history.first_turn()

    while not game.is_over() and turn < game_history.last_turn():
        opt_capture_square = game.opponent_move_results()
        if game.turn == color:
            player.handle_opponent_move_result(opt_capture_square is not None, opt_capture_square)

        sense_actions = game.sense_actions()
        move_actions = game.move_actions()

        sense = game_history.sense(turn)
        player_sense = player.choose_sense(sense_actions, move_actions, game.get_seconds_left())
        if game.turn == color and sense != player_sense:
            print('Warning: Sense action did not match history on turn {}. Using the sense action from history.'.format(
                turn))
        sense_result = game.sense(sense)
        if game.turn == color:
            player.handle_sense_result(sense_result)

        move = game_history.requested_move(turn)
        player_move = player.choose_move(move_actions, game.get_seconds_left())
        if game.turn == color and move != player_move:
            print('Warning: Move action did not match history on turn {}. Using the move action from history.'.format(
                turn))
        requested_move, taken_move, opt_enemy_capture_square = game.move(move)
        if game.turn == color:
            player.handle_move_result(requested_move, taken_move,
                                      opt_enemy_capture_square is not None, opt_enemy_capture_square)

        game.end_turn()
        turn = turn.next

    game.end()
    winner_color = game.get_winner_color()
    win_reason = game.get_win_reason()
    game_history = game.get_game_history()

    player.handle_game_end(winner_color, win_reason, game_history)


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('game_history_path', help='Path to game history JSON file.')
    parser.add_argument('bot_path', help='The path to bot source file or module.')
    parser.add_argument('color', default='random', choices=['white', 'black'],
                        help='The color of the player to playback.')
    args = parser.parse_args()

    game_history = GameHistory.from_file(args.game_history_path)
    white_bot_name, white_player_cls = load_player(args.bot_path)
    color = chess.WHITE if args.color == 'white' else chess.BLACK

    playback(game_history, white_player_cls(), color)


if __name__ == '__main__':
    main()



import random
from reconchess import *


class RandomBot(Player):
    def handle_game_start(self, color: Color, board: chess.Board, opponent_name: str):
        pass

    def handle_opponent_move_result(self, captured_my_piece: bool, capture_square: Optional[Square]):
        pass

    def choose_sense(self, sense_actions: List[Square], move_actions: List[chess.Move], seconds_left: float) -> \
            Optional[Square]:
        return random.choice(sense_actions)

    def handle_sense_result(self, sense_result: List[Tuple[Square, Optional[chess.Piece]]]):
        pass

    def choose_move(self, move_actions: List[chess.Move], seconds_left: float) -> Optional[chess.Move]:
        return random.choice(move_actions + [None])

    def handle_move_result(self, requested_move: Optional[chess.Move], taken_move: Optional[chess.Move],
                           captured_opponent_piece: bool, capture_square: Optional[Square]):
        pass

    def handle_game_end(self, winner_color: Optional[Color], win_reason: Optional[WinReason],
                        game_history: GameHistory):
        pass



import argparse
import requests
import multiprocessing
import time
import getpass
import traceback
from datetime import datetime
import reconchess
from reconchess import load_player, play_remote_game
import sys
import signal


class RBCServer:
    def __init__(self, server_url, auth):
        self.server_url = server_url
        self.invitations_url = '{}/api/invitations'.format(server_url)
        self.user_url = '{}/api/users'.format(server_url)
        self.me_url = '{}/api/users/me'.format(server_url)
        self.game_url = '{}/api/games'.format(server_url)
        self.session = requests.Session()
        self.session.auth = auth

    def _get(self, endpoint):
        response = self.session.get(endpoint)
        while response.status_code >= 500:
            time.sleep(0.5)
            response = self.session.get(endpoint)
        if response.status_code == 401:
            print('Authentication Error!')
            print(response.text)
            quit()
        return response.json()

    def _post(self, endpoint, json=None):
        response = self.session.post(endpoint, json=json)
        while response.status_code >= 500:
            time.sleep(0.5)
            response = self.session.post(endpoint, json=json)
        if response.status_code == 401:
            print('Authentication Error!')
            print(response.text)
            quit()
        return response.json()

    def get_reconchess_version(self):
        return self._get('{}/api/version'.format(self.server_url))['version']

    def set_max_games(self, max_games):
        self._post('{}/max_games'.format(self.me_url), json={'max_games': max_games})

    def get_active_users(self):
        return self._get('{}/'.format(self.user_url))['usernames']

    def send_invitation(self, opponent, color):
        return self._post('{}/'.format(self.invitations_url), json={
            'opponent': opponent,
            'color': color,
        })['game_id']

    def get_invitations(self):
        return self._get('{}/'.format(self.invitations_url))['invitations']

    def accept_invitation(self, invitation_id):
        return self._post('{}/{}'.format(self.invitations_url, invitation_id))['game_id']

    def finish_invitation(self, invitation_id):
        self._post('{}/{}/finish'.format(self.invitations_url, invitation_id))

    def error_resign(self, game_id):
        return self._post('{}/{}/error_resign'.format(self.game_url, game_id))

    def set_ranked(self, ranked):
        self._post('{}/ranked'.format(self.me_url), json={'ranked': ranked})

    def get_bot_version(self):
        return self._get('{}/version'.format(self.me_url))['version']

    def increment_version(self):
        self._post('{}/version'.format(self.me_url))


def accept_invitation_and_play(server_url, auth, invitation_id, bot_cls, finished):
    # make sure this process doesn't react to interrupt signals
    signal.signal(signal.SIGTERM, signal.SIG_IGN)
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    print('[{}] Accepting invitation {}.'.format(datetime.now(), invitation_id))

    server = RBCServer(server_url, auth)
    game_id = server.accept_invitation(invitation_id)

    print('[{}] Invitation {} accepted. Playing game {}.'.format(datetime.now(), invitation_id, game_id))

    try:
        play_remote_game(server_url, game_id, auth, bot_cls())
        print('[{}] Finished game {}'.format(datetime.now(), game_id))
    except:
        print('[{}] Fatal error in game {}:'.format(datetime.now(), game_id))
        traceback.print_exc()
        server.error_resign(game_id)
    finally:
        server.finish_invitation(invitation_id)
        finished.value = True


def check_package_version(server):
    server_version = server.get_reconchess_version()
    if server_version != reconchess.__version__:
        print('Your reconchess package is out of date!'.format(reconchess.__version__))
        print('Current:', 'reconchess v{}'.format(reconchess.__version__))
        print('Latest :', 'reconchess v{}'.format(server_version))
        print('Please run the command `pip install --upgrade reconchess` to upgrade to the latest.')
        quit()


def ranked_mode(server, keep_version):
    current_version = server.get_bot_version()

    # ask if they want to update their version number
    if current_version == 0:
        should_increment_version = True
    elif keep_version:
        should_increment_version = False
    else:
        should_increment_version = confirm('Is this a new version of your bot?')

    # compute the new version depending on their answer
    next_version = current_version + (1 if should_increment_version else 0)

    # make sure they want to participate with this info
    if not keep_version and not confirm(
            'Are you sure you want to participate in ranked matches as v{} (currently v{})?'.format(
                next_version, current_version)):
        quit()

    # update information on server
    if should_increment_version:
        server.increment_version()
    server.set_ranked(True)


def unranked_mode(server):
    server.set_ranked(False)


def listen_for_invitations(server, bot_cls, max_concurrent_games):
    connected = False
    process_by_invitation = {}
    finished_by_invitation = {}
    while True:
        try:
            # get unaccepted invitations
            invitations = server.get_invitations()

            # set max games on server if this is the first successful connection after being disconnected
            if not connected:
                print('[{}] Connected successfully to server!'.format(datetime.now()))
                connected = True
                server.set_max_games(max_concurrent_games)

            # filter out finished processes
            finished_invitations = []
            for invitation in process_by_invitation.keys():
                if not process_by_invitation[invitation].is_alive() or finished_by_invitation[invitation].value:
                    finished_invitations.append(invitation)
            for invitation in finished_invitations:
                print('[{}] Terminating process for invitation {}'.format(datetime.now(), invitation))
                process_by_invitation[invitation].terminate()
                del process_by_invitation[invitation]
                del finished_by_invitation[invitation]

            # accept invitations until we have #max_concurrent_games processes alive
            for invitation in invitations:
                # only accept the invitation if we have room and the invite doesn't have a process already
                if invitation not in process_by_invitation:
                    print('[{}] Received invitation {}.'.format(datetime.now(), invitation))

                    if len(process_by_invitation) < max_concurrent_games:
                        # start the process for playing a game
                        finished = multiprocessing.Value('b', False)
                        process = multiprocessing.Process(
                            target=accept_invitation_and_play,
                            args=(server.server_url, server.session.auth, invitation, bot_cls, finished))
                        process.start()

                        # store the process so we can check when it finishes
                        process_by_invitation[invitation] = process
                        finished_by_invitation[invitation] = finished
                    else:
                        print('[{}] Not enough game slots to play invitation {}.'.format(datetime.now(), invitation))

        except requests.RequestException as e:
            connected = False
            print('[{}] Failed to connect to server'.format(datetime.now()))
            print(e)
        except Exception:
            print("Error in invitation processing: ")
            traceback.print_exc()

        time.sleep(5)


def ask_for_username():
    return input('Username: ')


def ask_for_password():
    return getpass.getpass()


def ask_for_auth():
    return ask_for_username(), ask_for_password()


def confirm(prompt):
    while True:
        response = input(prompt + ' [y/n]')
        if len(response) == 0 or response.lower()[0] == 'y':
            return True
        elif len(response) > 0 and response.lower()[0] == 'n':
            return False
        else:
            print('Invalid input... please enter one of {"y", "n", "yes", "no", <Enter>, ...}')


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bot_path', help='Path to bot source or bot module name.')
    parser.add_argument('--username', default=None, help='Username for login. Enter with prompt if not specified.')
    parser.add_argument('--password', default=None, help='Password for login. Enter with prompt if not specified.')
    parser.add_argument('--server-url', default='https://rbc.jhuapl.edu', help='URL of the server.')
    parser.add_argument('--ranked', action='store_true', default=False,
                        help='Whether you want to play ranked matches.')
    parser.add_argument('--keep-version', action='store_true', default=False,
                        help='Force your ranked version to stay the same with no prompts.')
    parser.add_argument('--max-concurrent-games', type=int, default=4,
                        help='The maximum number of games to play at the same time.')
    args = parser.parse_args()

    bot_name, bot_cls = load_player(args.bot_path)

    username = ask_for_username() if args.username is None else args.username
    password = ask_for_password() if args.password is None else args.password
    auth = username, password

    server = RBCServer(args.server_url, auth)

    # verify we have the correct version of reconchess package
    check_package_version(server)

    def handle_term(signum, frame):
        print('[{}] Received terminate signal, waiting for games to finish and then exiting...'.format(datetime.now()))
        unranked_mode(server)
        sys.exit(0)

    signal.signal(signal.SIGINT, handle_term)
    signal.signal(signal.SIGTERM, handle_term)

    # tell the server whether we want to do ranked matches or not
    if args.ranked:
        ranked_mode(server, args.keep_version)
    else:
        unranked_mode(server)

    listen_for_invitations(server, bot_cls, args.max_concurrent_games)


if __name__ == '__main__':
    main()



import argparse
import requests
from reconchess.scripts.rc_connect import ask_for_username, ask_for_password


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--username', default=None, help='Name to register under. Enter with prompt if not specified.')
    parser.add_argument('--email', default=None, help='Email to associate with account. Enter with prompt if not specified.')
    parser.add_argument('--affiliation', default=None, help='Organization you are affiliated with.')
    parser.add_argument('--password', default=None, help='Password to use. Enter with prompt if not specified.')
    parser.add_argument('--server-url', default='https://rbc.jhuapl.edu', help='URL of the server.')
    args = parser.parse_args()

    try:
        requests.get('{}/api/users'.format(args.server_url))
    except:
        print('No server found at {} - it may not be available yet.'.format(args.server_url))
        quit()

    username = ask_for_username() if args.username is None else args.username
    email = input('Email: ') if args.email is None else args.email
    affiliation = input('Affiliation: ') if args.affiliation is None else args.affiliation
    password = ask_for_password() if args.password is None else args.password

    response = requests.post('{}/api/users/'.format(args.server_url), json={
        'username': username,
        'email': email,
        'affiliation': affiliation,
        'password': password,
    })

    if response.status_code == 200:
        username = response.json()['username']
        print('Successfully registered with username "{}".'.format(username))
    elif response.status_code == 409:
        print('Unsuccessful... "{}" is already in use! Choose a different name.'.format(args.username))
    else:
        print('Unsuccessful... returncode={}'.format(response.status_code))


if __name__ == '__main__':
    main()



import pkg_resources
import argparse
import contextlib
import chess
from reconchess import GameHistory

# block output from pygame
with contextlib.redirect_stdout(None):
    import pygame


class Button:
    BUTTON_RECT_COLOR = (238, 238, 238)
    BUTTON_TEXT_COLOR = (0, 0, 0)

    BUTTON_DISABLED_RECT_COLOR = (238, 238, 238)
    BUTTON_DISABLED_TEXT_COLOR = (170, 170, 170)

    BUTTON_HOVER_RECT_COLOR = (224, 224, 224)
    BUTTON_HOVER_TEXT_COLOR = (0, 0, 0)

    BUTTON_DOWN_RECT_COLOR = (200, 200, 200)
    BUTTON_DOWN_TEXT_COLOR = (0, 0, 0)

    def __init__(self, x, y, width, height, font, text, onclick=None):
        self.rect = (x, y, width, height)
        self.text = font.render(text, True, self.BUTTON_TEXT_COLOR)
        self.hover_text = font.render(text, True, self.BUTTON_HOVER_TEXT_COLOR)
        self.down_text = font.render(text, True, self.BUTTON_DOWN_TEXT_COLOR)
        self.disabled_text = font.render(text, True, self.BUTTON_DISABLED_TEXT_COLOR)

        text_width, text_height = font.size(text)
        self.text_position = (x - text_width / 2 + width / 2, y - text_height / 2 + height / 2)

        self.onclick = onclick
        self.enabled = True

        self.is_down = False

    def mouse_is_over(self):
        x, y = pygame.mouse.get_pos()
        return self.rect[0] < x < self.rect[0] + self.rect[2] and self.rect[1] < y < self.rect[1] + self.rect[3]

    def is_hovered(self):
        return pygame.mouse.get_focused() and self.mouse_is_over() and not bool(pygame.mouse.get_pressed()[0])

    def is_pressed(self):
        return pygame.mouse.get_focused() and self.mouse_is_over() and bool(pygame.mouse.get_pressed()[0])

    def update(self):
        released = self.is_down and not self.is_pressed()
        if self.enabled and self.onclick is not None and released:
            self.onclick()
        self.is_down = self.is_pressed()

    def draw(self, background):
        if not self.enabled:
            text = self.disabled_text
            rect_color = self.BUTTON_DISABLED_RECT_COLOR
        elif self.is_hovered():
            text = self.hover_text
            rect_color = self.BUTTON_HOVER_RECT_COLOR
        elif self.is_pressed():
            text = self.down_text
            rect_color = self.BUTTON_DOWN_RECT_COLOR
        else:
            text = self.text
            rect_color = self.BUTTON_RECT_COLOR

        pygame.draw.rect(background, rect_color, self.rect)
        background.blit(text, self.text_position)


class ReplayWindow:
    LIGHT_COLOR = (240, 217, 181)
    DARK_COLOR = (181, 136, 99)

    CAPTURE_COLOR = (255, 0, 0)

    LIGHT_SQUARES = list(chess.SquareSet(chess.BB_LIGHT_SQUARES))
    DARK_SQUARES = list(chess.SquareSet(chess.BB_DARK_SQUARES))

    def __init__(self, history: GameHistory):
        self.actions = []
        for turn in history.turns():
            if history.has_sense(turn):
                self.actions.append({
                    'phase': 'sense',
                    'turn_number': turn.turn_number,
                    'turn_color': turn.color,
                    'sense': history.sense(turn),
                    'sense_result': history.sense_result(turn),
                    'fen': history.truth_fen_before_move(turn) if history.is_first_turn(
                        turn) else history.truth_fen_after_move(turn.previous),
                })
            if history.has_move(turn):
                self.actions.append({
                    'phase': 'move',
                    'turn_number': turn.turn_number,
                    'turn_color': turn.color,
                    'requested_move': history.requested_move(turn),
                    'taken_move': history.taken_move(turn),
                    'capture_square': history.capture_square(turn),
                    'fen': history.truth_fen_after_move(turn),
                })

        self.board = chess.Board()
        self.action_index = None

        self.fps = 30
        self.square_size = 80
        self.width = self.square_size * 8
        self.height = self.square_size * 9

        self.image_by_piece = {}
        for color in chess.COLORS:
            for piece_type in chess.PIECE_TYPES:
                piece = chess.Piece(piece_type, color)

                img_path = 'res/{}/{}.png'.format(chess.COLOR_NAMES[color], piece.symbol())
                full_path = pkg_resources.resource_filename('reconchess', img_path)

                img = pygame.image.load(full_path)
                img = pygame.transform.scale(img, (self.square_size, self.square_size))
                self.image_by_piece[piece] = img

        pygame.init()
        pygame.display.set_caption('Recon Chess')
        pygame.display.set_icon(
            pygame.transform.scale(self.image_by_piece[chess.Piece(chess.KING, chess.WHITE)], (32, 32)))

        self.clock = pygame.time.Clock()
        self.screen = pygame.display.set_mode((self.width, self.height))
        self.background = pygame.Surface((self.screen.get_size()))

        self.perspective = chess.WHITE

        self.font = pygame.font.SysFont(pygame.font.get_default_font(), 30)
        self.buttons = []

        x, y = self.text_coords_below(chess.C1)
        self.buttons.append(Button(x, y, self.square_size / 2, self.square_size / 4, self.font, '<<',
                                   onclick=self.go_to_beginning))

        x, y = self.text_coords_below(chess.D1)
        self.buttons.append(Button(x, y, self.square_size / 2, self.square_size / 4, self.font, '<',
                                   onclick=self.go_backwards))

        x, y = self.text_coords_below(chess.E1)
        self.buttons.append(Button(x, y, self.square_size / 2, self.square_size / 4, self.font, '>',
                                   onclick=self.go_forwards))

        x, y = self.text_coords_below(chess.F1)
        self.buttons.append(Button(x, y, self.square_size / 2, self.square_size / 4, self.font, '>>',
                                   onclick=self.go_to_end))

        self.go_to_beginning()

    def go_to_beginning(self):
        self.action_index = None
        self.buttons[0].enabled = False
        self.buttons[1].enabled = False
        self.buttons[2].enabled = True
        self.buttons[3].enabled = True

    def go_to_end(self):
        self.action_index = len(self.actions) - 1
        self.buttons[0].enabled = True
        self.buttons[1].enabled = True
        self.buttons[2].enabled = False
        self.buttons[3].enabled = False

    def go_forwards(self):
        if self.action_index is None:
            self.action_index = 0
        else:
            self.action_index += 1
        self.buttons[0].enabled = True
        self.buttons[1].enabled = True
        self.buttons[2].enabled = self.action_index < len(self.actions) - 1
        self.buttons[3].enabled = self.action_index < len(self.actions) - 1

    def go_backwards(self):
        if self.action_index == 0:
            self.action_index = None
        else:
            self.action_index -= 1
        self.buttons[0].enabled = self.action_index is not None
        self.buttons[1].enabled = self.action_index is not None
        self.buttons[2].enabled = True
        self.buttons[3].enabled = True

    def coords_to_square(self, x, y):
        file = (x // self.square_size)
        if self.perspective == chess.WHITE:
            rank = ((self.height - y) // self.square_size)
        else:
            rank = (y // self.square_size)
        return chess.square(file, rank)

    def square_to_coords(self, square):
        rank = chess.square_rank(square)
        file = chess.square_file(square)
        x = file * self.square_size
        if self.perspective == chess.WHITE:
            y = (7 - rank) * self.square_size
        else:
            y = rank * self.square_size
        return x, y

    def square_rect(self, square):
        x, y = self.square_to_coords(square)
        return x, y, self.square_size, self.square_size

    def text_coords_below(self, square, horizontal_offset=0.25, vertical_offset=0.375):
        x, y = self.square_to_coords(square)
        x += horizontal_offset * self.square_size
        y += (1 + vertical_offset) * self.square_size
        return x, y

    def update(self):
        self.clock.tick(self.fps)
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                quit()

        for btn in self.buttons:
            btn.update()

    def draw(self):
        self.background.fill((238, 238, 238))

        self.draw_board()
        for btn in self.buttons:
            btn.draw(self.background)

        if self.action_index is not None:
            if self.actions[self.action_index]['phase'] == 'sense':
                self.draw_sense()
            else:
                self.draw_move()

        pygame.draw.line(self.background, (170, 170, 170), (0, self.square_size * 8),
                         (self.square_size * 8, self.square_size * 8))

        self.draw_turn_info()

        self.screen.blit(self.background, (0, 0))
        pygame.display.flip()

    def draw_board(self):
        if self.action_index is None:
            self.board.set_fen(chess.STARTING_FEN)
        else:
            self.board.set_fen(self.actions[self.action_index]['fen'])

        for square in chess.SQUARES:
            color = self.LIGHT_COLOR if square in self.LIGHT_SQUARES else self.DARK_COLOR
            pygame.draw.rect(self.background, color, self.square_rect(square))

            piece = self.board.piece_at(square)
            if piece is not None:
                image = self.image_by_piece[piece]
                self.background.blit(image, self.square_rect(square))

    def draw_sense(self):
        for square, opt_piece in self.actions[self.action_index]['sense_result']:
            self.draw_highlight(square, color=self.turn_color())

    def draw_move(self):
        requested_move = self.actions[self.action_index]['requested_move']
        taken_move = self.actions[self.action_index]['taken_move']
        capture_square = self.actions[self.action_index]['capture_square']

        if requested_move is not None and taken_move is None:
            self.draw_highlight(requested_move.from_square, color=(255, 0, 0))
            self.draw_highlight(requested_move.to_square, color=(255, 0, 0))
        elif requested_move is not None and taken_move is not None:
            self.draw_highlight(requested_move.from_square, color=self.turn_color())
            self.draw_highlight(requested_move.to_square, color=self.turn_color())
            self.draw_highlight(taken_move.from_square, color=self.turn_color())
            self.draw_highlight(taken_move.to_square, color=self.turn_color())

    def draw_turn_info(self):
        turn = '-' if self.action_index is None else (self.actions[self.action_index]['turn_number'] + 1)
        player = '-' if self.action_index is None else chess.COLOR_NAMES[self.actions[self.action_index]['turn_color']]

        text = 'Turn: {} / {}'.format(turn, self.actions[-1]['turn_number'] + 1)
        text_width, text_height = self.font.size(text)
        x, y = self.text_coords_below(chess.A1, vertical_offset=0.33)
        x, y = (x, y - text_height / 2 + .33)
        self.background.blit(self.font.render(text, True, (0, 0, 0)), (x, y))

        text = 'Player: {}'.format(player)
        text_width, text_height = self.font.size(text)
        x, y = self.text_coords_below(chess.A1, vertical_offset=0.66)
        x, y = (x, y - text_height / 2 + .33)
        self.background.blit(self.font.render(text, True, (0, 0, 0)), (x, y))

    def turn_color(self):
        return (255, 255, 255) if self.actions[self.action_index]['turn_color'] else (0, 0, 0)

    def draw_highlight(self, square, color=(255, 255, 0)):
        pygame.draw.rect(self.background, color, self.square_rect(square), 3)


def main():
    parser = argparse.ArgumentParser(description='Allows you to watch a saved match.')
    parser.add_argument('history_path', help='Path to saved Game History file.')
    args = parser.parse_args()

    history = GameHistory.from_file(args.history_path)
    if history.is_empty():
        print('Game History is empty.')
        quit()

    window = ReplayWindow(history)

    while True:
        window.update()
        window.draw()


if __name__ == '__main__':
    main()



import setuptools
import reconchess

with open('requirements.txt') as fp:
    requirements = fp.readlines()

with open('README.rst') as fp:
    long_description = fp.read()

setuptools.setup(
    name='reconchess',
    version=reconchess.__version__,
    author='Johns Hopkins University Applied Physics Laboratory',
    description='A python package for building Reconnaissance Chess players',
    long_description=long_description,
    long_description_content_typ='text/x-rst',
    url='https://github.com/reconnaissanceblindchess/reconchess',
    packages=['reconchess', 'reconchess.scripts', 'reconchess.bots'],
    package_data={'reconchess': ['res/white/*.png', 'res/black/*.png']},
    entry_points={
        'console_scripts': [
            'rc-bot-match=reconchess.scripts.rc_bot_match:main',
            'rc-play=reconchess.scripts.rc_play:main',
            'rc-replay=reconchess.scripts.rc_replay:main',
            'rc-playback=reconchess.scripts.rc_playback:main',
            'rc-register=reconchess.scripts.rc_register:main',
            'rc-connect=reconchess.scripts.rc_connect:main',
            'rc-play-on-server=reconchess.scripts.rc_play_on_server:main',
        ],
    },
    python_requires='>=3.5',
    install_requires=requirements,
    project_urls={
        'Documentation': 'https://reconchess.readthedocs.io/en/latest/index.html',
        'Source': 'https://github.com/reconnaissanceblindchess/reconchess',
        'Tracker': 'https://github.com/reconnaissanceblindchess/reconchess/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Games/Entertainment :: Board Games',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries',
    ],
)



from typing import List, Tuple, Optional, Type
from enum import Enum

Square = int
Color = bool
PieceType = int


class WinReason(Enum):
    """The reason the game ended"""

    KING_CAPTURE = 1
    """The game ended because one player captured the other's king."""

    TIMEOUT = 2
    """The game ended because one player ran out of time."""

    RESIGN = 3
    """The game ended because one player resigned."""

    TURN_LIMIT = 4
    """The game ended because it exceeded the full turn limit."""

    MOVE_LIMIT = 5
    """The game ended because it exceeded the reversible move limit."""
